from bs4 import BeautifulSoup
from collections import UserDict
from datetime import datetime
from dexy.exceptions import InternalDexyProblem
from dexy.exceptions import UserFeedback
from dexy.plugin import TemplatePlugin
from dexy.utils import levenshtein
from dexy.version import DEXY_VERSION
from pygments.styles import get_all_styles
import calendar
import dexy.commands
import dexy.commands.cite
import dexy.data
import dexy.exceptions
import dexy.plugin
import inflection
import inspect
import jinja2
import json
import markdown
import operator
import os
import pygments
import pygments.formatters
import random
import re
import time
import uuid
import xml.etree.ElementTree as ET

class Etree(TemplatePlugin):
    """
    Exposes element tree as ET.
    """
    aliases = ['etree']
    def run(self):
        return { 'ET' : ("The xml.etree.ElementTree module.", ET,) }

class Markdown(TemplatePlugin):
    """
    Exposes markdown.
    """
    aliases = ['md', 'markdown']

    def run(self):
        md = markdown.Markdown()
        h = "Function which converts markdown to HTML."
        return {
                'markdown' : (h, md.convert),
                'md' : (h, md.convert)
                }

class Uuid(TemplatePlugin):
    """
    Exposes the UUID module.
    """
    aliases = ['uuid']
    def run(self):
        return { 'uuid' : ("The Python uuid module. http://docs.python.org/2/library/uuid.html", uuid) }

class Time(TemplatePlugin):
    """
    Exposes time module.
    """
    aliases = ['time']
    def run(self):
        return { 'time' : ("The Python time module.", time) }

class Random(TemplatePlugin):
    """
    Exposes random module.
    """
    aliases = ['random']

    def shuffle(self, input_array):
        return sorted(input_array, key=lambda *args: random.random())

    def run(self):
        return { 'random' : ("The Python random module.", random),
                'shuffle' : ("Random shuffle not in place.", self.shuffle) }

class Operator(TemplatePlugin):
    """
    Exposes features of the operator module.
    """
    aliases = ['operator']
    keys = ['attrgetter', 'itemgetter']
    def run(self):
        d = {}
        for k in self.keys:
            fn = getattr(operator, k)
            d[k] = ("The %s method from Python's operator module." % k, fn)
        return d

class PrettyPrintHtml(TemplatePlugin):
    """
    Uses BeautifulSoup 4 to prettify HTML.
    """
    aliases = ['bs4']
    _settings = {
            'no-jinja-filter' : ['BeautifulSoup']
            }

    def prettify_html(self, html):
        soup = BeautifulSoup(str(html), 'html.parser')
        return soup.prettify()

    def run(self):
        return {
            'prettify_html' : ("Pretty-print HTML using BeautifulSoup", self.prettify_html),
            'BeautifulSoup' : ("The BeautifulSoup module.", BeautifulSoup)
            }

class LoadYaml(TemplatePlugin):
    """
    Loads YAML from a file.
    """
    aliases = ['loadyaml']

    def load_yaml(self, filename):
        import yaml
        with open(filename, 'rb') as f:
            return yaml.safe_load(f.read())

    def run(self):
        return {
                'load_yaml' : ("Safely load YAML from a file.", self.load_yaml,)
                }

class ParseYaml(TemplatePlugin):
    """
    Parse YAML from a string.
    """
    aliases = ['parseyaml']

    def parse_yaml(self, yamltext):
        import yaml
        return yaml.safe_load(str(yamltext))

    def run(self):
        return {
                'parse_yaml' : ("Safely load YAML from text.", self.parse_yaml,)
                }

class Debug(TemplatePlugin):
    """
    Adds debug() and throw() [a.k.a. raise()] methods to templates.
    """
    aliases = ['debug']

    def debug(self, debug_text, echo=True):
        if hasattr(self, 'filter_instance'):
            print("template debug from '%s': %s" % (self.filter_instance.key, debug_text))
        else:
            print("template debug: %s" % (debug_text))

        if echo:
            return debug_text
        else:
            return ""

    def throw(self, err_message):
        if hasattr(self, 'filter_instance'):
            raise UserFeedback("template throw from '%s': %s" % (self.filter_instance.key, err_message))
        else:
            raise UserFeedback("template throw: %s" % (err_message))

    def run(self):
        return {
                'debug' : ("A debugging method - prints content to command line stdout.", self.debug),
                'throw' : ("A debugging utility which raises exception with argument.", self.throw),
                'raise' : ("Alias for `throw`.", self.throw)
                }

class Bibtex(TemplatePlugin):
    """
    Produces a bibtex entry for dexy.
    """
    def run(self):
        return { 'dexy_bibtex' : dexy.commands.cite.bibtex_text() }

class Inflection(TemplatePlugin):
    """
    Exposes the inflection package for doing nice things with strings 
    """
    aliases = ['inflection']
    _settings = {
            'methods' : ("Methods of the inflection module to expose.",
                ['camelize', 'dasherize', 'humanize', 'ordinal',
                'ordinalize', 'parameterize', 'pluralize', 'singularize',
                'titleize', 'transliterate', 'underscore'])
            }

    def run(self):
        return dict((method, ("The %s method from Python inflection module." % method, getattr(inflection, method),)) for method in self.setting('methods'))

class StripJavadocHTML(TemplatePlugin):
    """
    Exposes javadoc2rst command which strips HTML tags from javadoc comments.
    """
    aliases = ['stripjavadochtml']
    _settings = {
            'escape' : ("Escape characters.", ['\\']),
            'remove' : ("Remove characters.", ['<p>', '<P>'])
            }

    def strip_javadoc_html(self, javadoc):
        for symbol in self.setting('escape'):
            javadoc = javadoc.replace(symbol, '\n')
        for word in self.setting('remove'):
            javadoc = javadoc.replace(word, "\\%s" % symbol)
        return javadoc

    def run(self):
        h = "Replace escape character with newlines and remove paragraph tags."
        return {
                'javadoc2rst' : (h, self.strip_javadoc_html),
                'strip_javadoc_html' : (h, self.strip_javadoc_html)
                }

class PrettyPrint(TemplatePlugin):
    """
    Exposes pprint (really pformat).
    """
    aliases = ['pp', 'pprint']

    def run(self):
        import pprint
        return {
            'pprint' : ("Pretty prints Python objects.", pprint.pformat,),
            'pformat' : ("Pretty prints Python objects.", pprint.pformat,)
         }

class PrettyPrintJson(TemplatePlugin):
    """
    Exposes ppjson command.
    """
    aliases = ['ppjson']

    def ppjson(self, json_string):
        return json.dumps(json.loads(json_string), sort_keys = True, indent = 4)

    def run(self):
        return {
            'ppjson' : ("Pretty prints valid JSON.", self.ppjson,)
         }

class ReplaceJinjaFilters(TemplatePlugin):
    """
    Replace some jinja filters so they call unicode() first.
    """
    aliases = ['replacejinjafilters']

    def do_indent(self, data, width=4, indentfirst=False):
        return jinja2.filters.do_indent(str(data), width, indentfirst)

    def run(self):
        return {
                'indent' : ("Jinja's indent function.", self.do_indent)
                }

class Assertions(TemplatePlugin):
    """
    Allow making assertions in documents.
    """
    aliases = ['assertions']

    def decorate_response(self, doc):
        if hasattr(self, 'filter_instance'):
            indicator = self.filter_instance.setting('assertion-passed-indicator')
        else:
            indicator = None

        if indicator:
            return str(doc) + indicator
        else:
            return doc

    def do_assert_equals(self, doc, expected):
        """
        Assert that input equals expected value.
        """
        assert str(doc) == expected, "input text did not equal '%s'" % expected
        return self.decorate_response(doc)

    def do_assert_contains(self, doc, contains):
        """
        Assert that input equals expected value.
        """
        assert contains in str(doc), "input text did not contain '%s'" % contains
        return self.decorate_response(doc)

    def do_assert_does_not_contain(self, doc, shouldnt_contain):
        """
        Assert that input equals expected value.
        """
        msg = "input text contained '%s'" % shouldnt_contain
        assert not shouldnt_contain in str(doc), msg
        return self.decorate_response(doc)

    def do_assert_startswith(self, doc, startswith):
        """
        Assert that the input starts with the specified value.
        """
        assert str(doc).startswith(startswith), "input text did not start with '%s'" % startswith
        return self.decorate_response(doc)

    def do_assert_matches(self, doc, regexp):
        """
        Assert that input matches the specified regular expressino.
        """
        assert re.match(regexp, str(doc)), "input text did not match regexp %s" % regexp
        return self.decorate_response(doc)

    def make_soup(self, doc):
        return BeautifulSoup(str(doc))

    def soup_select(self, doc, selector):
        soup = self.make_soup(doc)
        return soup.select(selector)

    def soup_select_unique(self, doc, selector):
        results = self.soup_select(doc, selector)

        n = len(results)
        if n == 0:
            msg = "no results found matching selector '%s'"
            msgargs = (selector,)
            raise AssertionError(msg % msgargs)

        elif n > 1:
            msg = "%s results found matching selector '%s', must be unique"
            msgargs = (n, selector,)
            raise AssertionError(msg % msgargs)

        return results[0]

    def do_assert_selector_text(self, doc, selector, expected_text):
        """
        Asserts that the contents of CSS selector matches the expected text.
        
        Leading/trailing whitespace is stripped before comparison.
        """
        element = self.soup_select_unique(doc, selector)
        err = "element '%s' did not contain '%s'" % (selector, expected_text)
        assert element.get_text().strip() == expected_text, err

    def run(self):
        methods = {}
        for name in dir(self):
            if name.startswith("do_"):
                method = getattr(self, name)
                docs = inspect.getdoc(method).splitlines()[0].strip()
                if not docs:
                    raise InternalDexyProblem("You must define docstring for %s" % name)
                methods[name.replace("do_", "")] = (docs, method)
        return methods

class Head(TemplatePlugin):
    """
    Provides a 'head' method.
    """
    aliases = ['head']

    def head(self, text, n=15):
        """
        Returns the first n lines of input string.
        """
        return "\n".join(str(text).splitlines()[0:n])

    def run(self):
        return {
                'head' : ("Returns the first n lines of input string.", self.head)
                }

class Tail(TemplatePlugin):
    """
    Provides a 'tail' method.
    """
    aliases = ['tail']

    def tail(self, text, n=15):
        """
        Returns the last n lines of input string.
        """
        return "\n".join(str(text).splitlines()[-n:])

    def run(self):
        return {
                'tail' : ("Returns the last n lines of inptu string.", self.tail)
                }

class RstCode(TemplatePlugin):
    """
    Indents code n spaces (defaults to 4) and wraps in .. code:: directive.
    """
    aliases = ['rstcode']

    def rstcode(self, text, n=4, language='python'):
        output = inspect.cleandoc("""
            .. code:: %s
            '   :class: highlight'
            """ % language)

        for line in str(text).splitlines():
            output += " " * n + line

        output += "\n"
        return output

class PythonDatetime(TemplatePlugin):
    """
    Exposes python datetime and calendar functions.
    """
    aliases = ['datetime', 'calendar']
    def run(self):
        today = datetime.today()
        month = today.month
        year = today.year
        cal = calendar.Calendar()
        caldates = list(cal.itermonthdates(year, month))

        return {
            "datetime" : ("The Python datetime module.", datetime),
            "calendar" : ("A Calendar instance from Python calendar module.", calendar),
            "caldates" : ("List of calendar dates in current month.", caldates),
            "cal" : ("Shortcut for `calendar`.", calendar),
            "today" : ("Result of datetime.today().", today),
            "month" : ("Current month.", month),
            "year" : ("Current year.", year)
            }

class DexyVersion(TemplatePlugin):
    """
    Exposes the current dexy version
    """
    aliases = ['dexyversion']
    def run(self):
        return { "DEXY_VERSION" : ("The active dexy version. Currently %s." % DEXY_VERSION, DEXY_VERSION) }

class SimpleJson(TemplatePlugin):
    """
    Exposes the json module.
    """
    aliases = ['json']
    def run(self):
        return { 'json' : ("The Python json module.", json,) }

class RegularExpressions(TemplatePlugin):
    """
    Exposes re_match and re_search.
    """
    aliases = ['regex']
    def run(self):
        return { 're' : ("The Python re module.", re,), }

class PythonBuiltins(TemplatePlugin):
    """
    Exposes python builtins.
    """
    aliases = ['builtins']
    # Intended to be all builtins that make sense to run within a document.
    PYTHON_BUILTINS = [abs, all, any, bin, bool, bytearray,
            callable, chr, complex, dict, dir, divmod, enumerate, filter,
            float, format, hex, id, int, isinstance, issubclass, iter, len,
            list, locals, map, hasattr, max, min, oct, ord, pow, range,
            repr, reversed, round, set, slice, sorted, str, sum, tuple,
            type, str, zip]

    def run(self):
        return dict((f.__name__, ("The python builtin function %s" % f.__name__, f,)) for f in self.PYTHON_BUILTINS)

class PygmentsStylesheet(TemplatePlugin):
    """
    Inserts pygments style codes.
    """
    aliases = ['pygments']

    # TODO rewrite this so it's a function rather than pre-generating all
    # of the stylesheets. Detect document format automatically.

    def generate_stylesheets(self):
        pygments_stylesheets = {}
        if hasattr(self, 'filter_instance') and 'pygments' in self.filter_instance.doc.args:
            formatter_args = self.filter_instance.doc.args['pygments']
        else:
            formatter_args = {}

        for style_name in get_all_styles():
            for formatter_class in [pygments.formatters.LatexFormatter, pygments.formatters.HtmlFormatter]:
                formatter_args['style'] = style_name
                pygments_formatter = formatter_class(**formatter_args)
                style_info = pygments_formatter.get_style_defs()

                for fn in pygments_formatter.filenames:
                    ext = fn.split(".")[1]
                    if ext == 'htm':
                        ext = 'css'
                    key = "%s.%s" % (style_name, ext)
                    pygments_stylesheets[key] = style_info

        return pygments_stylesheets

    def run(self):
        return {
            'pygments' : (
                "Dictionary of pygments stylesheets.",
                self.generate_stylesheets()
             )}

class PygmentsHighlight(TemplatePlugin):
    """
    Provides a 'highlight' function for applying syntax highlighting.
    """
    aliases = ['highlight']
    # TODO figure out default fmt based on document ext - document would need
    # to implement a "final_ext()" method

    def highlight(self, text, lexer_name, fmt='html', noclasses=False, style=None, lineanchors='l'):
        text = str(text)
        formatter_options = { "lineanchors" : lineanchors, "noclasses" : noclasses }
        if style is not None:
            formatter_options['style'] = style
        lexer = pygments.lexers.get_lexer_by_name(lexer_name)
        formatter = pygments.formatters.get_formatter_by_name(fmt, **formatter_options)
        return pygments.highlight(text, lexer, formatter)

    def run(self):
        return {
                'highlight' : ("Pygments syntax highlighter.", self.highlight),
                'pygmentize' : ("Pygments syntax highlighter.", self.highlight)
                }

class Subdirectories(TemplatePlugin):
    """
    Show subdirectories under this document.
    """
    aliases = ['subdirectories']
    def run(self):
        if hasattr(self.filter_instance, 'output_data'):
            # The directory containing the document to be processed.
            doc_dir = os.path.dirname(self.filter_instance.output_data.name)

            # Get a list of subdirectories under this document's directory.
            subdirectories = [d for d in sorted(os.listdir(os.path.join(os.curdir, doc_dir))) if os.path.isdir(os.path.join(os.curdir, doc_dir, d))]
            return {'subdirectories' : ("List of subdirectories of this document.", subdirectories)}
        else:
            return {'subdirectories' : ("List of subdirectories of this document.", [])}

class Variables(TemplatePlugin):
    """
    Allow users to set variables in document args which will be available to an individual document.
    """
    aliases = ['variables']
    def run(self):
        variables = {}
        variables.update(self.filter_instance.setting('variables'))
        variables.update(self.filter_instance.setting('vars'))

        formatted_variables = {}
        for k, v in variables.items():
            if isinstance(v, tuple):
                formatted_variables[k] = v
            elif isinstance(v, list) and len(v) == 2:
                formatted_variables[k] = v
            else:
                formatted_variables[k] = ("User-provided variable.", v)
        return formatted_variables

class Globals(TemplatePlugin):
    """
    Makes available the global variables specified on the dexy command line
    using the --globals option
    """
    aliases = ['globals']
    def run(self):
        raw_globals = self.filter_instance.doc.wrapper.globals
        env = {}
        for kvpair in raw_globals.split(","):
            if "=" in kvpair:
                k, v = kvpair.split("=")
                env[k] = ("Global variable %s" % k, v)
        return env

class Inputs(TemplatePlugin):
    """
    Populates the 'd' object.
    """
    aliases = ['inputs']

    def input_tasks(self):
        return self.filter_instance.doc.walk_input_docs()

    def run(self):
        input_docs = {}

        for doc in self.input_tasks():
            self.filter_instance.log_debug("adding input doc %s" % doc.key)
            input_docs[doc.key] = doc

        d = D(self.filter_instance.doc, input_docs)

        if hasattr(self.filter_instance, 'output_data'):
            output_data = self.filter_instance.output_data
        else:
            output_data = None

        return {
            'a' : ("Another way to reference 'd'. Deprecated.", d),
            'args' : ("The document args.", self.filter_instance.doc.args),
            'd' : ("The 'd' object.", d),
            'f' : ("The filter instance for this document.", self.filter_instance),
            's' : ("The data instance for this document.", output_data),
            'w' : ("The wrapper for the dexy run.", self.filter_instance.doc.wrapper)
            }

class D(UserDict):
    def __init__(self, doc, input_docs):
        self._artifact = doc
        self._parent_dir = doc.output_data().parent_dir()
        self._input_docs_dict = input_docs
        self._input_docs = list(input_docs.values())
        self._input_doc_keys = list(input_docs.keys())
        self._input_doc_names = [d.output_data().long_name() for d in self._input_docs]
        self._input_doc_titles = ["title:%s" % d.output_data().title() for d in self._input_docs]

        self._ref_cache = {}

    def keys(self):
        return self._input_doc_keys

    def items(self):
        return self._input_docs_dict.items()

    def key_or_name_index(self, ref):
        if ref in self._input_doc_keys:
            return self._input_doc_keys.index(ref)
        elif ref in self._input_doc_names:
            return self._input_doc_names.index(ref)

    def matching_keys(self, ref):
        return [(i, k) for (i, k) in enumerate(self._input_doc_keys) if k.startswith(ref)]

    def unique_matching_key(self, ref):
        """
        If the reference unambiguously identifies a single key, return it.
        """
        matching_keys = self.matching_keys(ref)
        if len(matching_keys) == 1:
            return matching_keys[0]

    def title_index(self, ref):
        if ref in self._input_doc_titles:
            return self._input_doc_titles[ref]

    def path_to(self, other):
        if self._parent_dir:
            return os.path.normpath(os.path.join(self._parent_dir, other))
        else:
            return other

    def __getitem__(self, ref):
        try:
            return self._ref_cache[ref]
        except KeyError:
            pass

        doc = None
        path_to_ref = None
        index = None

        if ref.startswith("/"):
            path_to_ref = ref.lstrip("/")
            index = self.key_or_name_index(path_to_ref)

        elif ref.startswith('title:'):
            index = self.title_index(ref)

        else:
            path_to_ref = self.path_to(ref)
            index = self.key_or_name_index(path_to_ref)

        if index is not None:
            doc = self._input_docs[index]
        else:
            matching_key = self.unique_matching_key(ref)
            if matching_key:
                doc = self._input_docs[matching_key[0]]

        if doc:
            # store this reference in cache for next time
            self._ref_cache[ref] = doc.output_data()
            return doc.output_data()
        else:
            msg = "No document named '%s'\nis available as an input to '%s'.\n"

            closest_match_lev = 15 # if more than this, not worth mentioning
            closest_match = None

            self._artifact.log_warn("Listing documents which are available:")
            for k in sorted(self._input_doc_keys):
                lev = levenshtein(k, path_to_ref)
                if lev < closest_match_lev:
                    closest_match = k
                    closest_match_lev = lev
                self._artifact.log_warn("  %s" % self.path_to(k))

            msg += "There are %s input documents available, their keys have been written to dexy's log.\n" % len(self._input_doc_keys)
           
            if closest_match:
                path_to_closest_match = self.path_to(closest_match)
                msg += "Did you mean '%s'?" % path_to_closest_match

            msgargs = (ref, self._artifact.key)
            raise dexy.exceptions.UserFeedback(msg % msgargs)



from dexy.doc import Doc
from tests.utils import assert_output
from tests.utils import runfilter
from tests.utils import wrap

def test_process_text_filter():
    assert_output("processtext", "hello", "Dexy processed the text 'hello'")

def test_process_method():
    assert_output("process", "hello", "Dexy processed the text 'hello'")

def test_process_method_manual_write():
    assert_output("processmanual", "hello", "Dexy processed the text 'hello'")

def test_process_method_with_dict():
    assert_output("processwithdict", "hello", {'1' : "Dexy processed the text 'hello'"})

def test_add_new_document():
    with runfilter("newdoc", "hello") as doc:
        assert doc.children[-1].key == "subdir/newfile.txt|processtext"
        assert str(doc.output_data()) == "we added a new file"

        assert "doc:subdir/example.txt|newdoc" in doc.wrapper.nodes
        assert "doc:subdir/newfile.txt|processtext" in doc.wrapper.nodes

def test_key_value_example():
    with wrap() as wrapper:
        doc = Doc(
                "hello.txt|keyvalueexample",
                wrapper,
                [],
                contents="hello"
                )

        wrapper.run_docs(doc)

        print(str(doc.output_data()))

        assert doc.output_data()["foo"] == "bar"
        assert str(doc.output_data()) == "KeyValue('hello.txt|keyvalueexample')"

def test_access_other_documents():
    with wrap() as wrapper:
        node = Doc("hello.txt|newdoc", wrapper, [], contents="hello")
        parent = Doc("test.txt|others",
                wrapper,
                [node],
                contents="hello"
                )
        wrapper.run_docs(parent)

        expected_items = [
            "Here is a list of previous docs in this tree (not including test.txt|others).",
            "newfile.txt|processtext (0 children, 0 inputs, length 33)",
            "hello.txt|newdoc (1 children, 0 inputs, length 19)"
            ]

        output = str(parent.output_data())

        for item in expected_items:
            assert item in output



from tests.utils import wrap
import dexy.filter

def test_filters_by_tag():
    tags_filters = dexy.filter.filters_by_tag()
    assert 'latex' in list(tags_filters.keys())

def test_filter_aliases_by_tag():
    first_expected_tag = 'R'
    first_actual = dexy.filter.filter_aliases_by_tag()[0][0]
    assert first_actual == first_expected_tag, first_actual

def test_filter_iter():
    for instance in dexy.filter.Filter:
        assert instance.setting('aliases')

def test_filter_args():
    with wrap() as wrapper:
        import dexy.node
        doc = dexy.node.Node.create_instance('doc',
                "hello.txt|filterargs",
                wrapper,
                [],
                contents="hello",
                foo="bar",
                filterargs={"abc" : 123, "foo" : "baz" }
                )

        wrapper.run_docs(doc)

        result = str(doc.output_data())

        assert "Here are the filter settings:" in result
        assert "abc: 123" in result
        assert "foo: baz" in result
        assert "foo: bar" in result




from dexy.filters.api import ApiFilter
import dexy.exceptions
import json
import xmlrpc.client

try:
    import mimetypes
    wp_aliases = ['wp', 'wordpress']
except UnicodeDecodeError:
    print("Unable to load mimetypes library. WordPressFilter will not work. See http://bugs.python.org/issue9291")
    mimetypes = None
    wp_aliases = []

class WordPressFilter(ApiFilter):
    """
    Posts to a WordPress blog.

    WordPress has a very confusing API setup since it implements its own API
    methods (under the wp namespace) and also supports the Blogger, metaWeblog
    and MovableType APIs. Unfortunately the wp namespace methods are
    incomplete, so you have to mix and match.

    Uses the WP XMLRPC API where possible:
    http://codex.wordpress.org/XML-RPC_wp

    Creating and editing blog posts uses the metaWeblog API:
    http://xmlrpc.scripting.com/metaWeblogApi.html

    If this filter is applied to a document with file extension in
    PAGE_CONTENT_EXTENSIONS (defined in ApiFilter class and inherited here)
    then the document will be uploaded to WordPress as a blog post.

    If not, then the document is assumed to be an image or other binary asset,
    and file upload will be used instead, so a new element will be added to the
    Media Library. If this is the case, then the URL is the resulting image is
    returned, so you can use that URL directly in your blog posts or other
    documents that need to link to the asset.

    IMPORTANT There is currently a frustrating bug in WP:
    http://core.trac.wordpress.org/ticket/17604
    which means that every time you run this filter, a *new* image asset will
    be created, even though we tell WordPress to overwrite the existing image
    of the same name. You will end up with dozens of copies of this image
    cluttering up your media library.

    For now, we recommend using an external site to host your images and
    assets, such as Amazon S3.
    """
    aliases = wp_aliases

    _settings = {
            'blog-id' : ("The wordpress blog id.", 0),
            'page-content-extensions' : ('', ['.md', '.txt', '.html']),
            'document-api-config-file' : 'wordpress.json',
            'api-key-name' : 'wordpress',
            'output-extensions' : ['.txt']
            }

    def docmd_create_keyfile(self):
        """
        Creates a key file for WordPress in the local directory.
        """
        self.create_keyfile("project-api-key-file")

    def api_url(self):
        base_url = self.read_param('url')
        if base_url.endswith("xmlrpc.php"):
            return base_url
        else:
            if not base_url.endswith("/"):
                base_url = "%s/" % base_url
            return "%sxmlrpc.php" % base_url

    def api(klass):
        if not hasattr(klass, "_api"):
            klass._api = xmlrpc.client.ServerProxy(klass.api_url())
        return klass._api

    def docmd_list_methods(klass):
        """
        List API methods exposed by WordPress API.
        """
        for method in sorted(klass.api().system.listMethods()):
            print(method)

    def docmd_list_categories(self):
        """
        List available blog post categories.
        """
        username = self.read_param('username')
        password = self.read_param('password')
        headers = ['categoryName']
        print("\t".join(headers))
        for category_info in self.api().wp.getCategories(self.setting('blog-id'), username, password):
            print("\t".join(category_info[h] for h in headers))

    def upload_page_content(self):
        input_text = str(self.input_data)
        document_config = self.read_document_config()

        document_config['description'] = input_text
        post_id = document_config.get('postid')
        publish = document_config.get('publish', False)

        for key, value in document_config.items():
            if not key == "description":
                self.log_debug("%s: %s" % (key, value))

        if post_id:
            self.log_debug("Making editPost API call.")
            self.api().metaWeblog.editPost(
                    post_id,
                    self.read_param('username'),
                    self.read_param('password'),
                    document_config,
                    publish
                    )
        else:
            self.log_debug("Making newPost API call.")
            post_id = self.api().metaWeblog.newPost(
                    self.setting('blog-id'),
                    self.read_param('username'),
                    self.read_param('password'),
                    document_config,
                    publish
                    )
            document_config['postid'] = post_id

        self.log_debug("Making getPost API call.")
        post_info = self.api().metaWeblog.getPost(
                post_id,
                self.read_param('username'),
                self.read_param('password')
                )

        for key, value in post_info.items():
            if key in ('date_modified_gmt', 'dateCreated', 'date_modified', 'date_created_gmt',):
                post_info[key] = value.value

            if not key == "description":
                self.log_debug("%s: %s" % (key, value))

        del document_config['description']
        document_config['publish'] = publish
        self.save_document_config(document_config)

        if publish:
            self.output_data.set_data(post_info['permaLink'])
        else:
            self.output_data.set_data(json.dumps(post_info))

    def upload_image_content(self):
        with open(self.input_data.storage.data_file(), 'rb') as f:
            image_base_64 = xmlrpc.client.Binary(f.read())

            upload_file = {
                     'name' : self.work_input_filename(),
                     'type' : mimetypes.types_map[self.prev_ext],
                     'bits' : image_base_64,
                     'overwrite' : 'true'
                     }

            upload_result = self.api().wp.uploadFile(
                     self.setting('blog-id'),
                     self.read_param('username'),
                     self.read_param('password'),
                     upload_file
                     )

            self.log_debug("wordpress upload results: %s" % upload_result)
            url = upload_result['url']
            self.log_debug("uploaded %s to %s" % (self.key, url))

        self.output_data.set_data(url)

    def process(self):
        try:
            if self.prev_ext in self.setting('page-content-extensions'):
                self.upload_page_content()
            else:
                self.upload_image_content()

        except xmlrpc.client.Fault as e:
            raise dexy.exceptions.UserFeedback(str(e))



from dexy.reporter import Reporter
from jinja2 import Environment
from jinja2 import FileSystemLoader
import json
import os

class D3Graph(Reporter):
    """
    Generates a tree graph using d3.js
    """
    aliases = ['d3tree']
    _settings = {
            'dir' : 'd3tree',
            'in-cache-dir' : True,
            'filename' : 'd3tree.js',
            "run-for-wrapper-states" : ["ran", "checked", "error"]
            }

    def node_info(self, node):
        inputs_and_children = node.inputs + node.children
        return {
                'name' : node.key,
                'contents': [self.node_info(n) for n in inputs_and_children]
                }

    def run(self, wrapper):
        self.wrapper = wrapper
        self.remove_reports_dir(wrapper)
        self.copy_template_files()

        env_data = self.run_plugins()

        graphs = {
                "name" : "",
                "contents" : [self.node_info(node) for node in wrapper.roots]
                }

        env_data['graph_json'] = json.dumps(graphs)

        env = Environment()
        env.loader = FileSystemLoader(os.path.dirname(__file__))
        template = env.get_template('template.js')

        template.stream(env_data).dump(self.report_file(), encoding="utf-8")




from dexy.doc import Doc
from dexy.node import Node
from dexy.node import PatternNode
from dexy.wrapper import Wrapper
from tests.utils import wrap
import dexy.doc
import dexy.node
import os
import time

def test_create_node():
    with wrap() as wrapper:
        node = dexy.node.Node.create_instance(
                "doc",
                "foo.txt",
                wrapper,
                [],
                # kwargs
                foo='bar',
                contents="these are contents"
                )

        assert node.__class__ == dexy.doc.Doc
        assert node.args['foo'] == 'bar'
        assert node.wrapper == wrapper
        assert node.inputs == []
        assert len(node.hashid) == 32

def test_node_arg_caching():
    with wrap() as wrapper:
        wrapper.nodes = {}
        node = dexy.node.Node("foo", wrapper, [], foo='bar', baz=123)
        wrapper.add_node(node)

        assert node.hashid == 'acbd18db4cc2f85cedef654fccc4a4d8'
        assert node.args['foo'] == 'bar'
        assert node.args['baz'] == 123
        assert node.sorted_arg_string() == '[["baz", 123], ["foo", "bar"]]'

        assert os.path.exists(wrapper.artifacts_dir)
        assert not os.path.exists(wrapper.node_argstrings_filename())
        wrapper.save_node_argstrings()
        assert os.path.exists(wrapper.node_argstrings_filename())
        wrapper.load_node_argstrings()
        assert not node.check_args_changed()

        node.args['baz'] = 456
        assert node.check_args_changed()
        wrapper.save_node_argstrings()
        wrapper.load_node_argstrings()
        assert not node.check_args_changed()

SCRIPT_YAML = """
script:scriptnode:
    - start.sh|shint
    - middle.sh|shint
    - end.sh|shint
"""

def test_script_node_caching__slow():
    with wrap():
        with open("start.sh", "w") as f:
            f.write("pwd")

        with open("middle.sh", "w") as f:
            f.write("echo `time`")

        with open("end.sh", "w") as f:
            f.write("echo 'done'")

        with open("dexy.yaml", "w") as f:
            f.write(SCRIPT_YAML)

        wrapper1 = Wrapper(log_level="DEBUG")
        wrapper1.run_from_new()

        for node in list(wrapper1.nodes.values()):
            assert node.state == 'ran'

        wrapper2 = Wrapper()
        wrapper2.run_from_new()

        for node in list(wrapper2.nodes.values()):
            assert node.state == 'consolidated'

        time.sleep(1.1)
        with open("middle.sh", "w") as f:
            f.write("echo 'new'")

        wrapper3 = Wrapper()
        wrapper3.run_from_new()

        for node in list(wrapper1.nodes.values()):
            assert node.state == 'ran'

# TODO mock out os.stat to get different mtimes without having to sleep?

def test_node_caching__slow():
    with wrap() as wrapper:
        with open("hello.py", "w") as f:
            f.write("print(1+2)\n")

        with open("doc.txt", "w") as f:
            f.write("1 + 1 = {{ d['hello.py|py'] }}")

        wrapper = Wrapper(log_level='DEBUG')
        hello_py = Doc("hello.py|py", wrapper)
        doc_txt = Doc("doc.txt|jinja",
                wrapper,
                [hello_py]
                )

        wrapper.run_docs(doc_txt)

        assert str(doc_txt.output_data()) == "1 + 1 = 3\n"
        assert str(hello_py.output_data()) == "3\n"

        assert hello_py.state == 'ran'
        assert doc_txt.state == 'ran'

        wrapper = Wrapper(log_level='DEBUG')
        hello_py = Doc("hello.py|py", wrapper)
        doc_txt = Doc("doc.txt|jinja",
                wrapper,
                [hello_py]
                )
        wrapper.run_docs(doc_txt)

        assert hello_py.state == 'consolidated'
        assert doc_txt.state == 'consolidated'

        time.sleep(1.1)
        with open("doc.txt", "w") as f:
            f.write("1 + 1 = {{ d['hello.py|py'] }}\n")

        wrapper = Wrapper(log_level='DEBUG')
        hello_py = Doc("hello.py|py", wrapper)
        doc_txt = Doc("doc.txt|jinja",
                wrapper,
                [hello_py]
                )
        wrapper.run_docs(doc_txt)

        assert hello_py.state == 'consolidated'
        assert doc_txt.state == 'ran'

        time.sleep(1.1)
        with open("hello.py", "w") as f:
            f.write("print(1+1)\n")

        wrapper = Wrapper(log_level='DEBUG')
        hello_py = Doc("hello.py|py", wrapper)
        doc_txt = Doc("doc.txt|jinja",
                wrapper,
                [hello_py]
                )
        wrapper.run_docs(doc_txt)

        assert hello_py.state == 'ran'
        assert doc_txt.state == 'ran'

def test_node_init_with_inputs():
    with wrap() as wrapper:
        node = Node("foo.txt",
                wrapper,
                [Node("bar.txt", wrapper)]
                )
        assert node.key == "foo.txt"
        assert node.inputs[0].key == "bar.txt"

        expected = {
                0 : "bar.txt",
                1 : "foo.txt" 
            }
    
        for i, n in enumerate(node.walk_inputs()):
            assert expected[i] == n.key

def test_doc_node_populate():
    with wrap() as wrapper:
        node = Node.create_instance(
                'doc', "foo.txt", wrapper,
                [], contents='foo')

        assert node.key_with_class() == "doc:foo.txt"

def test_doc_node_with_filters():
    with wrap() as wrapper:
        node = Node.create_instance('doc',
                "foo.txt|outputabc", wrapper, [], contents='foo')
        assert node.key_with_class() == "doc:foo.txt|outputabc"

def test_pattern_node():
    with wrap() as wrapper:
        with open("foo.txt", "w") as f:
            f.write("foo!")

        with open("bar.txt", "w") as f:
            f.write("bar!")

        wrapper = Wrapper(log_level='DEBUG')
        wrapper.to_valid()

        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        node = PatternNode("*.txt", 
                wrapper,
                [],
                foo="bar")
        assert node.args['foo'] == 'bar'
        wrapper.run_docs(node)
        assert len(node.children) == 2

        for child in node.children:
            assert child.__class__.__name__ == "Doc"
            assert child.args['foo'] == 'bar'
            assert child.key_with_class() in ["doc:foo.txt", "doc:bar.txt"]
            assert child.filters == []

def test_pattern_node_multiple_filters():
    with wrap() as wrapper:
        with open("foo.txt", "w") as f:
            f.write("foo!")

        wrapper = Wrapper(log_level='DEBUG')
        wrapper.to_valid()

        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        node = PatternNode("*.txt|dexy|dexy|dexy", wrapper=wrapper)
        doc = node.children[0]
        assert doc.key == "foo.txt|dexy|dexy|dexy"
        assert doc.filter_aliases == ['dexy', 'dexy', 'dexy']
        assert doc.parent == node

def test_pattern_node_one_filter():
    with wrap() as wrapper:
        with open("foo.txt", "w") as f:
            f.write("foo!")

        wrapper = Wrapper(log_level='DEBUG')
        wrapper.to_valid()

        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        node = PatternNode("*.txt|dexy", wrapper=wrapper)
        doc = node.children[0]
        assert doc.key == "foo.txt|dexy"
        assert doc.filter_aliases == ['dexy']
        assert doc.parent == node



from dexy.doc import Doc
from tests.utils import wrap
import json

REGETRON_INPUT_1 = "hello\n"
REGETRON_INPUT_2 = """\
this is some text
9
nine
this is 100 mixed text and numbers
"""

def test_regetron_filter():
    with wrap() as wrapper:
        wrapper.debug = False
        node = Doc("example.regex|regetron",
                wrapper,
                [
                    Doc("input1.txt",
                        wrapper,
                        [],
                        contents=REGETRON_INPUT_1),
                    Doc("input2.txt",
                        wrapper,
                        [],
                        contents=REGETRON_INPUT_2)
                    ],
                contents="^[a-z\s]+$"
                )

        wrapper.run_docs(node)
        
        if not wrapper.state == 'error':
            assert str(node.output_data()['input1.txt']) == """\
> ^[a-z\s]+$
0000: hello
> 

"""
            assert str(node.output_data()['input2.txt']) == """\
> ^[a-z\s]+$
0000: this is some text
0002: nine
> 

"""

def test_used_filter():
    with wrap() as wrapper:
        node = Doc("input.txt|used",
                wrapper,
                [
                    Doc("example.sed",
                        wrapper,
                        [],
                        contents="s/e/E/g")
                    ],
                contents="hello")

        wrapper.run_docs(node)
        assert str(node.output_data()) == "hEllo"

def test_sed_filter_single_simple_input_file():
    with wrap() as wrapper:
        node = Doc("example.sed|sed",
                wrapper,
                [
                    Doc("input.txt",
                        wrapper,
                        [],
                        contents="hello")
                    ],
                contents="s/e/E/g")

        wrapper.run_docs(node)
        assert str(node.output_data()) == "hEllo\n"

def test_sed_filter_single_input_file_with_sections():
    contents = json.loads("""[{},
    { "name" : "foo", "contents" : "hello" },
    { "name" : "bar", "contents" : "telephone" }
    ]""")

    with wrap() as wrapper:
        node = Doc("example.sed|sed",
                wrapper,
                [
                    Doc("input.txt",
                        wrapper,
                        [],
                        contents=contents,
                        data_class='sectioned'
                        )
                        ],
                contents="s/e/E/g")

        wrapper.run_docs(node)
        assert str(node.output_data()['foo']) == 'hEllo\n'
        assert str(node.output_data()['bar']) == 'tElEphonE\n'

def test_sed_filter_multiple_inputs():
    with wrap() as wrapper:
        node = Doc("example.sed|sed",
                wrapper,
                inputs = [
                    Doc("foo.txt",
                        wrapper,
                        [],
                        contents='hello'),
                    Doc("bar.txt",
                        wrapper,
                        [],
                        contents='telephone')
                    ],
                contents="s/e/E/g")

        wrapper.run_docs(node)
        assert str(node.output_data()['foo.txt']) == 'hEllo\n'
        assert str(node.output_data()['bar.txt']) == 'tElEphonE\n'



from dexy.utils import defaults
from dexy.commands.utils import dummy_wrapper
from dexy.parser import AbstractSyntaxTree
from dexy.parser import Parser

def parsers_command():
    wrapper = dummy_wrapper()
    ast = AbstractSyntaxTree(wrapper)

    processed_aliases = set()

    for alias in sorted(Parser.plugins):
        if alias in processed_aliases:
            continue

        parser = Parser.create_instance(alias, ast, wrapper)

        for alias in parser.aliases:
            processed_aliases.add(alias)

        print("%s Parser" % parser.__class__.__name__)
        print('')
        print(parser.setting('help'))
        print('')
        print("aliases:")
        for alias in parser.aliases:
            print("  %s" % alias)
        print('')

    print("Default parsers are: " + defaults['parsers'])
    print('')
    print("Dexy will only look for config files to parse in the root directory")
    print("of your project unless --recurse is specified.")
    print('')



from tests.plugins.test_templating_filters import run_jinja_filter
from dexy.doc import Doc
from tests.utils import wrap
import inspect

def test_assert_equals():
    assert str(run_jinja_filter("{{ 'foo' | assert_equals('foo') }}")) == 'foo'

def test_assert_equals_invalid():
    try:
        str(run_jinja_filter("{{ 'foo' | assert_equals('bar') }}"))
        raise Exception("should raise AssertionError")
    except AssertionError as e:
        assert str(e) == "input text did not equal 'bar'"

def test_assert_contains():
    assert str(run_jinja_filter("{{ 'foo bar' | assert_contains('foo') }}")) == 'foo bar'

def test_assert_contains_invalid():
    try:
        str(run_jinja_filter("{{ 'foo bar' | assert_contains('baz') }}"))
        raise Exception("should raise AssertionError")
    except AssertionError as e:
        assert str(e) == "input text did not contain 'baz'"

def test_assert_does_not_contain():
    assert str(run_jinja_filter("{{ 'foo bar' | assert_does_not_contain('baz') }}")) == 'foo bar'

def test_assert_does_not_contain_invalid():
    try:
        str(run_jinja_filter("{{ 'foo bar baz' | assert_does_not_contain('baz') }}"))
        raise Exception("should raise AssertionError")
    except AssertionError as e:
        assert str(e) == "input text contained 'baz'"

def test_assert_startswith():
    assert str(run_jinja_filter("{{ 'foo bar' | assert_startswith('foo') }}")) == 'foo bar'

def test_assert_startswith_invalid():
    try:
        str(run_jinja_filter("{{ 'foo bar' | assert_startswith('bar') }}"))
        raise Exception("should raise AssertionError")
    except AssertionError as e:
        assert str(e) == "input text did not start with 'bar'"

def test_assert_matches():
    assert str(run_jinja_filter("{{ 'foo bar baz' | assert_matches('^foo') }}")) == 'foo bar baz'

def test_assert_matches_invalid():
    try:
        str(run_jinja_filter("{{ 'foo bar' | assert_matches('^baz') }}"))
        raise Exception("should raise AssertionError")
    except AssertionError as e:
        assert str(e) == "input text did not match regexp ^baz"

def test_assert_selector():
    with wrap() as wrapper:
        node = Doc("hello.txt|jinja",
                wrapper,
                [
                    Doc("input.html",
                        wrapper,
                        [],
                        contents = inspect.cleandoc("""
                        <div id="foo">
                        This is contents of foo div.
                        </div>
                        """
                        ))
                    ],
                contents = "{{ d['input.html'] | assert_selector_text('#foo', 'This is contents of foo div.') }}"
                )
        wrapper.run_docs(node)

def test_assert_selector_invalid():
    with wrap() as wrapper:
        node = Doc("hello.txt|jinja",
                wrapper,
                [
                    Doc("input.html",
                        wrapper,
                        [],
                        contents = inspect.cleandoc("""
                        <div id="foo">
                        This is contents of foo div.
                        </div>
                        """
                        ))
                    ],
                contents = "{{ d['input.html'] | assert_selector_text('#foo', 'Not right.') }}"
                )

        try:
            wrapper.run_docs(node)
            raise Exception("should raise AssertionError")
        except AssertionError as e:
            assert str(e) == "element '#foo' did not contain 'Not right.'"



from dexy.exceptions import UserFeedback, InternalDexyProblem
from dexy.filter import DexyFilter
import os
import tempfile
import json

try:
    import pygit2
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

def repo_from_path(path):
    """
    Initializes a pygit Repository instance from a local repo at 'path'
    """
    repo = pygit2.init_repository(path, False)
    if repo.is_empty:
        raise UserFeedback("no git repository was found at '%s'" % path)
    return repo, None

def repo_from_url(url, remote_name="origin"):
    """
    Initializes a pygit Repository instance from a remote repo at 'url'
    """
    tempdir = tempfile.mkdtemp() # TODO move to .dexy/persistent/tempdir
    repo = pygit2.init_repository(tempdir, False)
    remote = repo.create_remote(remote_name, url)
    remote.fetch()
    return repo, remote

def generate_commit_info(commit):
    # Currently this is diffing with nothing, so the 'diff' is a full printout
    # of the current contents of the repo.
    diff = commit.tree.diff_to_tree()

    patches = []
    for i, patch in enumerate(diff):
        patches.append({
            'old-file-path' : patch.old_file_path,
            'new-file-path' : patch.new_file_path,
            'hunks' : [ { 
                'old-start' : hunk.old_start,
                'old-lines' : hunk.old_lines,
                'new-start' : hunk.new_start,
                'new-lines' : hunk.new_lines,
                'lines' : hunk.lines
                } for hunk in patch.hunks]
            })

    commit_info = {
            'author-name' : commit.author.name,
            'author-email' : commit.author.email,
            'message' : commit.message.strip(),
            'hex' : commit.hex,
            'patch' : diff.patch.encode('ascii', 'ignore'),
            'patches' : json.dumps(patches).encode('ascii', 'ignore')
            }

    return commit_info

class GitBase(DexyFilter):
    """
    Base class for various git-related filters.
    """
    aliases = []
    _settings = {
            'reference' : ("The reference to use.", None),
            'revision' : ("The revision to use, see 'man gitrevisions'.", None),
            'url-prefixes' : ("""Tuple of strings which mean the specified repo
                should be treated as a URL.""", ("http", "git"))
            }

    def is_active(self):
        return AVAILABLE

    def is_url(self, input_text):
        """
        Should the input text be treated as the URL of a remote git repo?
        """
        return input_text.startswith(self.setting('url-prefixes'))

    def reference(self, repo):
        if self.setting('reference'):
            return repo.lookup_reference(self.setting('reference'))
        else:
            refs = repo.listall_references()
            return repo.lookup_reference(refs[0])

    def revision(self, repo):
        if self.setting('revision'):
            return repo.revparse_single(self.setting('revision'))

    def work(self, repo, remote, ref):
        return "do stuff here in subclass"

    def process_text(self, input_text):
        if self.is_url(input_text):
            repo, remote = repo_from_url(input_text)
        else:
            repo, remote = repo_from_path(input_text)

        ref = self.reference(repo)

        # TODO capture GitError class
        return self.work(repo, remote, ref)

class GitBaseKeyValue(GitBase):
    """
    A filter using key-value storage to manage content from a git repo.
    """
    _settings = {
            'data-type' : 'keyvalue',
            'output-extensions' : ['.sqlite3', '.json']
            }

    def process(self):
        input_text = str(self.input_data)
        if self.is_url(input_text):
            repo, remote = repo_from_url(input_text)
        else:
            repo, remote = repo_from_path(input_text)

        ref = self.reference(repo)

        # TODO capture GitError class
        self.work(repo, remote, ref)

        self.output_data.save()

class Git(GitBase):
    """
    What should be default?
    """
    aliases = ['git']

    def work(self, repo, remote, ref):
        return "done"

class GitRepo(GitBase):
    """
    Adds all files in a repo to the project tree as additional documents.

    Files can be filtered to limit which ones are added.
    """
    aliases = ['repo']

    def work(self, repo, remote, ref):
        parent_dir = self.output_data.parent_dir()

        print(dir(ref))
        commit = repo[ref.target]
        tree = commit.tree

        def process_tree(tree, add_to_dir):
            for entry in tree:
                obj = repo[entry.oid]
                if obj.__class__.__name__ == 'Blob':
                    doc_key = os.path.join(add_to_dir, entry.name)
                    self.add_doc(doc_key, obj.data)
                elif obj.__class__.__name__ == 'Tree':
                    process_tree(obj, os.path.join(parent_dir, entry.name))
                else:
                    raise InternalDexyProblem(obj.__class__.__name__)

        process_tree(tree, parent_dir)
        # TODO return something more meaningful like a list of the files added.
        # doesn't matter that much since this repo is used for its side effects
        return "done"

class GitCommit(GitBaseKeyValue):
    """
    Returns key-value store information for the most recent commit, or the
    specified revision.
    """
    aliases = ['gitcommit']

    def work(self, repo, remote, ref):
        commit = repo[ref.target]
        commit_info = generate_commit_info(commit)

        for k, v in commit_info.items():
            self.output_data.append(k, v)

class GitLog(GitBase):
    """
    Returns a simple commit log for the specified repository.
    """
    aliases = ['gitlog']
    _settings = {
            'output-extensions' : ['.txt']
            }

    def work(self, repo, remote, ref):
        log = ""
        for commit in repo.walk(ref.target, pygit2.GIT_SORT_TIME):
            log += "%s: %s\n" % (commit.hex, commit.message.strip())
        return log



from dexy.commands.utils import init_wrapper
from dexy.utils import defaults

def reset_command(
        __cli_options=False,
        artifactsdir=defaults['artifacts_dir'], # Where dexy should store working files.
        logdir=defaults['log_dir'] # DEPRECATED
        ):
    """
    Clean out the contents of dexy's cache and reports directories.
    """
    wrapper = init_wrapper(locals())
    wrapper.remove_dexy_dirs()
    wrapper.remove_reports_dirs(keep_empty_dir=True)
    wrapper.create_dexy_dirs()

def build_reset_parser(parser):
    parser.set_defaults(cmd=reset_command)
    parser.add_argument('--artifacts_dir', defaults['artifacts_dir'])

def cleanup_command(
        __cli_options=False,
        artifactsdir=defaults['artifacts_dir'], # Where dexy should store working files.
        logdir=defaults['log_dir'], # DEPRECATED
        reports=True # Whether directories generated by reports should also be removed.
        ):
    """
    Remove the directories which dexy created, including working directories
    and reports.
    """
    wrapper = init_wrapper(locals())
    wrapper.remove_dexy_dirs()
    wrapper.remove_reports_dirs(reports)

def setup_command(
        __cli_options=False,
        artifactsdir=defaults['artifacts_dir'], # Where dexy should store working files.
        **kwargs):
    """
    Create the directories dexy needs to run.
    """
    wrapper = init_wrapper(locals())
    wrapper.create_dexy_dirs()




from dexy.utils import defaults
from dexy.utils import file_exists
from dexy.utils import parse_json
from dexy.utils import parse_yaml
from textwrap import TextWrapper
from textwrap import dedent
from inspect import cleandoc
import dexy.wrapper
import logging
import os
import yaml

class NullHandler(logging.Handler):
    def emit(self, record):
        pass

RENAME_PARAMS = {
        'artifactsdir' : 'artifacts_dir',
        'conf' : 'config_file',
        'dbalias' : 'db_alias',
        'dbfile' : 'db_file',
        'disabletests' : 'disable_tests',
        'dryrun' : 'dry_run',
        'excludealso' : 'exclude_also',
        'ignore' : 'ignore_nonzero_exit',
        'logfile' : 'log_file',
        'logformat' : 'log_format',
        'loglevel' : 'log_level',
        'logdir' : 'log_dir',
        'nocache' : 'dont_use_cache',
        'outputroot' : 'output_root'
        }

def default_config():
    wrapper = dexy.wrapper.Wrapper()
    conf = wrapper.__dict__.copy()

    for k in list(conf.keys()):
        if not k in list(defaults.keys()):
            del conf[k]

    reverse_rename = dict((v,k) for k, v in RENAME_PARAMS.items())
    for k in list(conf.keys()):
        renamed_key = reverse_rename.get(k, k)
        if renamed_key != k:
            conf[renamed_key] = conf[k]
            del conf[k]

    return conf

def rename_params(kwargs):
    renamed_args = {}
    for k, v in kwargs.items():
        renamed_key = RENAME_PARAMS.get(k, k)
        renamed_args[renamed_key] = v
    return renamed_args

def skip_params(kwargs):
    ok_params = {}
    for k, v in kwargs.items():
        if k in list(defaults.keys()):
            ok_params[k] = v
    return ok_params

def config_args(modargs):
    cliargs = modargs.get("__cli_options", {})
    kwargs = modargs.copy()

    config_file = modargs.get('conf', dexy.utils.defaults['config_file'])

    # Update from config file
    if file_exists(config_file):
        with open(config_file, "r") as f:
            if config_file.endswith(".conf"):
                try:
                    conf_args = parse_yaml(f.read())
                except dexy.exceptions.UserFeedback as yaml_exception:
                    try:
                        conf_args = parse_json(f.read())
                    except dexy.exceptions.UserFeedback as json_exception:
                        print("--------------------------------------------------")
                        print("Tried to parse YAML:")
                        print(yaml_exception)
                        print("--------------------------------------------------")
                        print("Tried to parse JSON:")
                        print(json_exception)
                        print("--------------------------------------------------")
                        raise dexy.exceptions.UserFeedback("Unable to parse config file '%s' as YAML or as JSON." % config_file)

            elif config_file.endswith(".yaml"):
                conf_args = parse_yaml(f.read())
            elif config_file.endswith(".json"):
                conf_args = parse_json(f.read())
            else:
                raise dexy.exceptions.UserFeedback("Don't know how to load config from '%s'" % config_file)
            if conf_args:
                # TODO raise error if 
                kwargs.update(conf_args)

    if cliargs: # cliargs may be False
        for k in list(cliargs.keys()):
            try:
                kwargs[k] = modargs[k]
            except KeyError:
                msg = "This command does not take a '--%s' argument." % k
                raise dexy.exceptions.UserFeedback(msg)

    # TODO allow updating from env variables, e.g. DEXY_ARTIFACTS_DIR

    return kwargs

def import_plugins_from_local_yaml_file(import_target):
    if os.path.exists(import_target):
        with open(import_target, 'rb') as f:
            yaml_content = yaml.safe_load(f.read())

        for alias, info_dict in yaml_content.items():
            if ":" in alias:
                prefix, alias = alias.split(":")
            else:
                prefix = 'filter'

            plugin_classes = dict((plugin_class.__name__.lower(), plugin_class)
                    for plugin_class in dexy.plugin.Plugin.__subclasses__())

            if not prefix in plugin_classes:
                msg = "'%s' not found, available aliases are %s"
                args = (prefix, ", ".join(list(plugin_classes.keys())))
                raise dexy.exceptions.UserFeedback(msg % args)

            cls = plugin_classes[prefix]

            if alias in cls.plugins:
                existing_plugin = cls.plugins[alias]
                plugin_settings = existing_plugin[1]
                plugin_settings.update(info_dict)
                cls.plugins[alias] = (existing_plugin[0], plugin_settings)
            else:
                cls.register_plugins_from_dict({alias : info_dict})

    else:
        # Don't raise exception if default files don't exist.
        if not import_target in defaults['plugins']:
            msg = "Could not find YAML file named '%s'" % import_target
            raise dexy.exceptions.UserFeedback(msg)

def import_plugins_from_local_python_file(import_target):
    if os.path.exists(import_target):
        import imp
        imp.load_source("custom_plugins", import_target)
    else:
        # Don't raise exception if default files don't exist.
        if not import_target in ('dexyplugin.py', 'dexyplugins.py',):
            msg = "Could not find python file named '%s'" % import_target
            raise dexy.exceptions.UserFeedback(msg)

def import_plugins_from_python_package(import_target):
    try:
        __import__(import_target)
    except ImportError:
        msg = "Could not find installed python package named '%s'" % import_target
        raise dexy.exceptions.UserFeedback(msg)

def import_extra_plugins(kwargs):
    if kwargs.get('plugins'):
        for import_target in kwargs.get('plugins').split():
            if import_target.endswith('.yaml'):
                import_plugins_from_local_yaml_file(import_target)
            elif import_target.endswith('.py'):
                import_plugins_from_local_python_file(import_target)
            else:
                import_plugins_from_python_package(import_target)

def init_wrapper(modargs, apply_defaults=False):
    if apply_defaults:
        modargs_with_defaults = defaults
        modargs_with_defaults.update(modargs)
    else:
        modargs_with_defaults = modargs
    
    kwargs = config_args(modargs_with_defaults)
    import_extra_plugins(kwargs)
    kwargs = rename_params(kwargs)
    kwargs = skip_params(kwargs)
    return dexy.wrapper.Wrapper(**kwargs)

def template_text(template):
    for wrapper in template.dexy(True):
        man_doc_key = 'doc:dexy.rst|jinja|rst2man'
        if man_doc_key in wrapper.nodes:
            man_doc = wrapper.nodes[man_doc_key].output_data().storage.data_file()

            import subprocess
            proc = subprocess.Popen(
                       ["man", man_doc],
                       stdout=subprocess.PIPE,
                       stderr=subprocess.STDOUT
                   )
            stdout, stderr = proc.communicate()
            return stdout
        else:
            return "no example found"

def dummy_wrapper():
    wrapper = dexy.wrapper.Wrapper()
    wrapper.log = logging.getLogger('dexy')
    wrapper.log.addHandler(NullHandler())
    wrapper.filemap = {}
    return wrapper

def rewrap_text(text, spaces=None, **kwargs):
    wrapper_args = {
            }

    if spaces:
        wrapper_args['initial_indent'] = ' ' * spaces
        wrapper_args['subsequent_indent'] = ' ' * spaces

    wrapper_args.update(kwargs)

    wrapper = TextWrapper(**wrapper_args)
    return wrapper.fill(cleandoc(text))

def indent_text(text, spaces):
    indented = []
    indent = " " * spaces
    for line in dedent(text).splitlines():
        indented.append("%s%s" % (indent, line))
    return "\n".join(indented)

def print_indented(text, spaces=0):
    print(indent_text(text, spaces))

def print_rewrapped(text, spaces=0):
    print(rewrap_text(text, spaces))



from dexy.exceptions import UserFeedback
from dexy.filters.api import ApiFilter

import urllib
import re


class WebSequenceDiagrams(ApiFilter):
    """
    Dexy filter to use WebSequenceDiagrams.com
    """
    aliases = ["wsd"]
    _settings = {
        "version": ("Version of the WebSequenceDiagrams API to address", "1"),
        "style": ("Style to use, e.g. 'patent' or 'napkin'", "default"),
        "key": ("Your WebSequenceDiagrams API Key if you have one", None),
        "api-url": "http://www.websequencediagrams.com/",
        "output-extensions": ['.png', '.svg', '.img', '.pdf'],
        "input-extensions": ['.wsd'],
        "output": True,
        'api-key-name': "wsd",
        "added-in-version": "1.0.13",

    }
    _unset = ['api-username', 'api-password']

    def process(self):
        request = {}
        settings = self.setting_values()
        request["message"] = self.input_data
        request["style"] = settings["style"]
        request["apiVersion"] = settings["version"]
        key = None
        try:
            key = settings["key"] or self.read_param("key")
        except KeyError:
            pass
        if key:
            request["apikey"] = key
            self.log_debug("Using key to access WebSequenceDiagrams")
        else:
            self.log_debug("Accessing WebSequenceDiagrams without API key")

        resource = urllib.parse.urlencode(request)
        self.log_debug("Fetching from WebSequenceDiagrams")
        response = urllib.urlopen(settings["api-url"]+"index.php", resource)
        line = response.readline()
        response.close()

        expr = re.compile("(\?(img|pdf|png|svg)=[a-zA-Z0-9]+)")
        match = expr.search(line)

        if match is None:
            self.log_error("Problem fetching from WebSequenceDiagrams")
            raise UserFeedback("Invalid response from server.")

        self.log_debug("Writing WebSequenceDiagrams result to {0}".format(
            self.output_filepath()))

        response = urllib.urlopen(settings["api-url"] + match.group(0))
        self.output_data.set_data(response.read())
        self.output_data.save()
        response.close()



from dexy.doc import Doc
from tests.utils import wrap
import dexy.data
import dexy.exceptions
import os

def test_sectioned_data_setitem_delitem():
    with wrap() as wrapper:
        contents=[
                {},
                {
                    "name" : "Welcome",
                    "contents" : "This is the first section."
                }
            ]

        doc = Doc("hello.txt",
                wrapper,
                [],
                data_type="sectioned",
                contents=contents
                )

        wrapper.run_docs(doc)
        data = doc.output_data()

        assert data.alias == 'sectioned'
        assert len(data) == 1

        # Add a new section
        data["Conclusions"] = "This is the final section."

        assert len(data) == 2
        
        assert str(data['Welcome']) == "This is the first section."
        assert str(data["Conclusions"]) == "This is the final section."

        # Modify an existing section
        data["Welcome"] = "This is the initial section."

        assert len(data) == 2

        assert str(data['Welcome']) == "This is the initial section."
        assert str(data["Conclusions"]) == "This is the final section."

        del data["Conclusions"]

        assert len(data) == 1
        assert list(data.keys()) == ["Welcome"]

def test_generic_data_unicode():
    with wrap() as wrapper:
        doc = Doc("hello.txt",
                wrapper,
                [],
                contents="\u2042 we know\n"
                )

        wrapper.run_docs(doc)
        data = doc.output_data()

        assert data.alias == 'generic'
        assert str(data) == "\u2042 we know\n"

        assert isinstance(str(data), str)

def test_generic_data_stores_string():
    with wrap() as wrapper:
        doc = Doc("hello.txt",
                wrapper,
                [],
                contents="hello"
                )

        wrapper.run_docs(doc)
        data = doc.output_data()

        assert data.alias == 'generic'
        assert data._data == "hello"

def test_sectioned_data_stores_list_of_dicts():
    with wrap() as wrapper:
        contents=[
                {},
                {
                    "name" : "Welcome",
                    "contents" : "This is the first section."
                }
            ]

        doc = Doc("hello.txt",
                wrapper,
                [],
                data_type="sectioned",
                contents=contents
                )

        wrapper.run_docs(doc)
        data = doc.output_data()

        assert data.alias == 'sectioned'
        assert data._data == contents
        assert data['Welcome']['contents'] == "This is the first section."
        assert data[0]['contents'] == "This is the first section."

def test_keyvalue_data_stores_dict():
    with wrap() as wrapper:
        doc = Doc("hello.json",
                wrapper,
                [],
                data_type="keyvalue",
                contents="dummy contents"
                )

        wrapper.run_docs(doc)
        data = doc.output_data()

        assert data.alias == 'keyvalue'
        assert list(data.keys()) == []

        data.append("foo", 123)
        data.append("bar", 456)

        assert sorted(data.keys()) == ["bar", "foo"]

def test_canonical_name():
    with wrap() as wrapper:
        doc = Doc("hello.txt",
                wrapper,
                [],
                contents="hello",
                output_name="yello.abc")

        wrapper.run_docs(doc)
        assert doc.output_data().output_name() == "yello.abc"
        wrapper.report()
        assert os.path.exists(os.path.join('output', 'yello.abc'))

def test_attempt_write_outside_project_root():
    with wrap() as wrapper:
        try:
            doc = Doc("../../example.txt",
                wrapper,
                [],
                contents = "hello")
            doc.setup()
            doc.setup_datas()
            assert False, 'should raise UserFeedback'
        except dexy.exceptions.UserFeedback as e:
            print(str(e))
            assert 'trying to write' in str(e)

def test_key_value_data():
    with wrap() as wrapper:
        settings = {
                'canonical-name' : 'doc.json',
                'storage-type' : 'json'
                }
        data = dexy.data.KeyValue("doc.json", ".json", "doc.json", settings, wrapper)
        data.setup_storage()

        assert not data._data
        assert data.storage._data == {}

        # We use the append method to add key-value pairs.
        data.append('foo', 'bar')
        assert len(list(data.keys())) == 1

        assert data.value('foo') == 'bar'
        assert data.storage['foo'] == 'bar'

def test_key_value_data_sqlite():
    with wrap() as wrapper:
        wrapper.to_walked()
        wrapper.to_checked()

        settings = {
                'canonical-name' : 'doc.sqlite3'
                }

        data = dexy.data.KeyValue("doc.sqlite3", ".sqlite3", "abc000", settings, wrapper)
        data.setup_storage()
        data.storage.connect()

        data.append('foo', 'bar')
        assert len(list(data.keys())) == 1

        assert data.value('foo') == 'bar'
        assert ["%s: %s" % (k, v) for k, v in data.items()][0] == "foo: bar"

def test_generic_data():
    with wrap() as wrapper:
        wrapper.to_walked()
        wrapper.to_checked()

        CONTENTS = "contents go here"

        # Create a GenericData object
        settings = {
                'canonical-name' : 'doc.txt'
                }
        data = dexy.data.Generic("doc.txt", ".txt", "abc000", settings, wrapper)
        data.setup_storage()

        # Assign some text contents
        data._data = CONTENTS
        assert data.has_data()
        assert not data.is_cached(True)

        # Save data to disk
        data.save()
        assert data.has_data()
        assert data.is_cached(True)
        assert data.filesize(True) > 10

        # Clear data from memory
        data._data = None

        # Load it again from disk
        data.load_data(True)
        assert data._data == CONTENTS

        assert data.as_text() == CONTENTS

def test_init_data():
    with wrap() as wrapper:
        settings = {
                'canonical-name' : 'doc.abc'
                }
        data = dexy.data.Generic("doc.txt", ".abc", "def123", settings, wrapper)
        data.setup_storage()

        assert data.key == "doc.txt"
        assert data.name == "doc.abc"
        assert data.ext == ".abc"
        assert data.storage_key == "def123"

        assert not data.has_data()
        assert not data.is_cached()



from dexy.doc import Doc
from dexy.exceptions import UserFeedback
from dexy.filters.id import lexer as id_lexer
from dexy.filters.id import parser as id_parser
from dexy.filters.id import start_new_section, token_info
from tests.utils import TEST_DATA_DIR
from tests.utils import wrap
import os

def test_force_text():
    with wrap() as wrapper:
        node = Doc("example.py|idio|t",
                wrapper,
                [],
                contents="print 'hello'\n")

        wrapper.run_docs(node)
        assert str(node.output_data()) == "print 'hello'\n"

def setup_parser():
    with wrap() as wrapper:
        id_parser.outputdir = wrapper.log_dir
        id_parser.errorlog = wrapper.log
        id_lexer.outputdir = wrapper.log_dir
        id_lexer.errorlog = wrapper.log
        id_lexer.remove_leading = True
        id_parser.write_tables = False
        _lexer = id_lexer.clone()
        _lexer.sections = []
        _lexer.level = 0
        start_new_section(_lexer, 0, 0, _lexer.level)
        yield(id_parser, _lexer)

def parse(text):
    for id_parser, _lexer in setup_parser():
        id_parser.parse(text, lexer=_lexer)
        return _lexer.sections

def tokens(text):
    for id_parser, _lexer in setup_parser():
        return token_info(text, _lexer)

def test_parse_code():
    output = parse("foo\n")
    assert output[0]['contents'] == 'foo\n'

def test_parse_oldstyle_comments():
    for comment in ('###', '///', ';;;', '%%%'):
        text = "%s @export foo\nfoo\n" % comment
        output = parse(text)
        assert output[0]['name'] == 'foo'
        assert output[0]['contents'] == 'foo\n'

def test_parse_comments():
    for comment in ('###', '///', ';;;', '%%%'):
        text = "%s 'foo-bar'\nfoo\n" % comment
        output = parse(text)
        assert output[0]['name'] == 'foo-bar'
        assert output[0]['contents'] == 'foo\n'

def test_parse_closed_style_sections():
    comments = (
        "/*** @export foo1*/\n",
        "/*** @export foo1 */\n",
        "/*** @section foo1 */\n",
        "/*** @export 'foo1'*/\n",
        "/*** @export 'foo1' */\n",
        "/*** @export 'foo1' python*/\n",
        "/*** @export 'foo1' python */\n",
        """/*** @export "foo1" css */\n""",
        """/*** @export "foo1" python*/\n""",
        "<!-- @export foo1 -->\n",
        "<!-- @section foo1 -->\n"
        "<!-- section foo1 -->\n"
        "<!-- section 'foo1' -->\n"
        )

    for text in comments:
        output = parse(text)
        assert output[0]['contents'] == ''
        assert output[0]['name'] == 'foo1'

def test_parse_closed_style_end():
    comments = (
        "foo\n/*** @end */\nbar\n",
        "foo\n<!-- @end -->\nbar\n",
        "foo\n<!-- section 'end' -->\nbar\n"
        )
    for text in comments:
        output = parse(text)
        assert output[0]['contents'] == 'foo\n'
        assert output[1]['contents'] == 'bar\n'
        assert output[0]['name'] == '1'
        assert output[1]['name'] == '2'

def test_parse_closed_falsestart():
    comments = (
        "<!-- @bob -->\n",
        "/*** @bob */\n"
        )

    for text in comments:
        output = parse(text)
        assert output[0]['contents'] == text

def test_ignore_faux_comment():
    for comment in ('#', '/', '%', '##%', '//#', '%#%', '##', '//', '%%',
            '///', "foo;", "//;", ";#;"):
        text = "  %s foo bar\nfoo\n" % comment
        output = parse(text)
        assert output[0]['contents'] == text

def test_accidental_comment_in_string():
    for comment in ('###",', '%%%",', '"###",', '"%%%"', '"### (%%%)",'):
        text = "foo bar %s\n" % comment
        output = parse(text)
        assert output[0]['contents'] == text

def test_more_accidental_comments():
    for comment in ('###",', '%%%",', '"###",', '"%%%"', '"### (%%%)",'):
        text = "   %s foo bar %s\n" % (comment, comment)
        output = parse(text)
        print("INPUT IS", text)
        print("OUTPUT IS", output[0]['contents'])
        assert output[0]['contents'] == text

def test_malformatted_comment_does_not_throw_error():
    for comment in ('###', '///', '%%%'):
        # There should be no space in this style of section name
        text = "%s 'foo bar'\nfoo\n" % comment
        output = parse(text)
        assert output[0]['contents'] == text

def test_idio_invalid_input():
    with wrap() as wrapper:
        wrapper.debug = False
        doc = Doc("hello.py|idio",
                wrapper, [],
                contents="### @ ")
        wrapper.run_docs(doc)

def test_multiple_sections():
    with wrap() as wrapper:
        src = """
### @export "vars"
x = 6
y = 7

### @export "multiply"
x*y

"""
        doc = Doc("example.py|idio",
                wrapper,
                [],
                contents=src)

        wrapper.run_docs(doc)
        assert list(doc.output_data().keys()) == ['1', 'vars', 'multiply']

def uest_force_latex():
    with wrap() as wrapper:
        doc = Doc("example.py|idio|l",
                wrapper,
                [],
                contents="print 'hello'\n")

        wrapper.run_docs(doc)

        assert "begin{Verbatim}" in str(doc.output_data())

def test_parse_docutils_latex():
    with open(os.path.join(TEST_DATA_DIR, "doc.tex"), "r") as f:
        latex = f.read()
    parse(latex)

def test_parse_php_mixed_tags():
    with open(os.path.join(TEST_DATA_DIR, "example.php"), "r") as f:
        php = f.read()

    output = parse(php)
    section_names = [info['name'] for info in output]

    assert "head" in section_names
    assert "assign-variables" in section_names
    assert "compare" in section_names
    assert "display-variables" in section_names



from dexy.commands.utils import default_config
from dexy.utils import defaults
from dexy.utils import file_exists
import dexy.exceptions
import inspect
import json
import os
import yaml

def conf_command(
        conf=defaults['config_file'], # name of config file to write to
        p=False # whether to print to stdout rather than write to file
        ):
    """
    Write a config file containing dexy's defaults.
    """
    if file_exists(conf) and not p:
        print(inspect.cleandoc("""Config file %s already exists,
        will print conf to stdout instead...""" % conf))
        p = True

    config = default_config()

    # Can't specify config file name in config file.
    del config['conf']

    yaml_help = inspect.cleandoc("""# YAML config file for dexy.
        # You can delete any lines you don't wish to customize.
        # Options are same as command line options,
        # for more info run 'dexy help -on dexy'.
        """)

    if p:
        print(yaml.dump(config, default_flow_style=False))
    else:
        with open(conf, "w") as f:
            if conf.endswith(".yaml") or conf.endswith(".conf"):
                f.write(yaml_help)
                f.write(os.linesep)
                f.write(yaml.dump(config, default_flow_style=False))
            elif conf.endswith(".json"):
                json.dump(config, f, sort_keys=True, indent=4)
            else:
                msg = "Don't know how to write config file '%s'"
                raise dexy.exceptions.UserFeedback(msg % conf)

        print("Config file has been written to '%s'" % conf)



from dexy.exceptions import InternalDexyProblem
from dexy.exceptions import UserFeedback
from dexy.exceptions import InactivePlugin
from dexy.filters.process import SubprocessFilter
import re
import os

try:
    import pexpect
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class DexyEOFException(UserFeedback):
    pass

class PexpectReplFilter(SubprocessFilter):
    """
    Use pexpect to retrieve output line-by-line based on detecting prompts.
    """
    _settings = {
            'trim-prompt' : ("The closing prompt to be trimmed off.", '>>>'),
            'send-line-ending' : ("Line ending to transmit at the end of each input line.", "\n"),
            'line-ending' : ("The line ending returned by REPL.", "\n"),
            'save-vars-to-json-cmd' : ("Command to be run to save variables to a JSON file.", None),
            'ps1' : ('PS1', None),
            'ps2' : ('PS2', None),
            'ps3' : ('PS3', None),
            'ps4' : ('PS4', None),
            'term' : ('TERM', 'dumb'),
            'initial-prompt' : ("The initial prompt the REPL will display.", None),
            'prompt' : ("Single prompt to match exactly.", None),
            'prompts' : ("List of possible prompts to match exactly.", ['>>>', '...']),
            'prompt-regex' : ("A prompt regex to match.", None),
            'strip-regex' : ("Regex to strip", None),
            'data-type' : 'sectioned',
            'allow-match-prompt-without-newline' : ("Whether to require a newline before prompt.", False),
            }

    def is_active(klass):
        return AVAILABLE

    def prompt_search_terms(self):
        """
        Search first for the prompt (or prompts) following a line ending.
        Also optionally allow matching the prompt with no preceding line ending.
        """
        prompt_regex = self.setting('prompt-regex')
        prompt = self.setting('prompt')

        if prompt_regex:
            prompts = [prompt_regex]
        elif prompt:
            prompts = [prompt]
        else:
            prompts = self.setting('prompts')

        if self.setting('allow-match-prompt-without-newline'):
            return ["%s%s" % (self.setting('line-ending'), p) for p in prompts] + prompts
        else:
            return ["%s%s" % (self.setting('line-ending'), p) for p in prompts]

    def lines_for_section(self, section_text):
        """
        Take the section text and split it into lines which will be sent to the
        T
        differently, or if you don't want the extra newline at the end.
        """
        return section_text.splitlines() + ["\n"]

    def strip_trailing_prompts(self, section_transcript):
        lines = section_transcript.splitlines()
        while len(lines) > 0 and re.match("^\s*(%s)\s*$|^\s*$" % self.setting('trim-prompt'), lines[-1]):
            lines = lines[0:-1]
        return self.setting('line-ending').join(lines)

    def strip_newlines(self, line):
        return line.replace(" \r", "")

    def section_output(self):
        """
        Runs the code in sections and returns an iterator so we can do custom stuff.
        """
        input_sections = self.input_data.items()

        # If we want to automatically record values of local variables in the
        # script we are running, we add a section at the end of script
        do_record_vars = self.setting('record-vars')
        if do_record_vars:
            if not self.setting('save-vars-to-json-cmd'):
                raise UserFeedback("You specified record-vars but this option isn't available since SAVE_VARS_TO_JSON_CMD is not set for this filter.")

            section_text = self.setting('save-vars-to-json-cmd') % self.input_data.basename()
            self.log_debug("Adding save-vars-to-json-cmd code:\n%s" % section_text)
            input_sections.append(('dexy--save-vars', section_text))
            if not self.setting('add-new-files'):
                docstr = self._instance_settings['add-new-files'][0]
                self._instance_settings['add-new-files'] = (docstr, ".json")

        search_terms = self.prompt_search_terms()

        env = self.setup_env()

        if self.setting('ps1'):
            ps1 = self.setting('ps1')
            self.log_debug("Setting PS1 to %s" % ps1)
            env['PS1'] = ps1

        if self.setting('ps2'):
            ps2 = self.setting('PS2')
            self.log_debug("Setting PS2 to %s" % ps2)
            env['PS2'] = ps2

        if self.setting('ps3'):
            ps3 = self.arg_value('PS3')
            self.log_debug("Setting PS3 to %s" % ps3)
            env['PS3'] = ps3

        if self.setting('ps4'):
            ps4 = self.arg_value('PS4')
            self.log_debug("Setting PS4 to %s" % ps4)
            env['PS4'] = ps4

        env['TERM'] = self.setting('term')

        timeout = self.setup_timeout()
        initial_timeout = self.setup_initial_timeout()

        self.log_debug("timeout set to '%s'" % timeout)

        if self.setting('use-wd'):
            wd = self.parent_work_dir()
        else:
            wd = os.getcwd()

        executable = self.setting('executable')
        self.log_debug("about to spawn new process '%s' in '%s'" % (executable, wd))

        # Spawn the process
        try:
            proc = pexpect.spawn(
                    executable,
                    cwd=wd,
                    env=env)
        except pexpect.ExceptionPexpect as e:
            if "The command was not found" in str(e):
                raise InactivePlugin(self)
            else:
                raise

        self.log_debug("Capturing initial prompt...")
        initial_prompt = self.setting('initial-prompt')
        try:
            if initial_prompt:
                proc.expect(initial_prompt, timeout=initial_timeout)
            elif self.setting('prompt-regex'):
                proc.expect(search_terms, timeout=initial_timeout)
            else:
                proc.expect_exact(search_terms, timeout=initial_timeout)

        except pexpect.TIMEOUT:
            if self.setting('initial-prompt'):
                match = self.setting('initial-prompt')
            else:
                match = search_terms

            msg = "%s failed at matching initial prompt within %s seconds. " % (self.__class__.__name__, initial_timeout)
            msg += "Received '%s', tried to match with '%s'" % (proc.before, match)
            msg += "\nExact characters received:\n"
            for i, c in enumerate(str(proc.before)):
                msg += "chr %02d: %s\n" % (i, ord(c))
            msg += "The developer might need to set a longer initial prompt timeout or the regexp may be wrong."
            raise InternalDexyProblem(msg)

        start = proc.before.decode('utf-8') + proc.after.decode('utf-8')

        self.log_debug("Initial prompt captured!")

        for section_key, section_text in input_sections:
            section_transcript = start
            start = ""

            lines = self.lines_for_section(section_text)
            for l in lines:
                self.log_debug("Sending '%s'" % l)
                section_transcript += start
                proc.send(l.rstrip() + self.setting('send-line-ending'))
                try:
                    if self.setting('prompt-regex'):
                        proc.expect(search_terms, timeout=timeout)
                    else:
                        proc.expect_exact(search_terms, timeout=timeout)

                    self.log_debug("Received '%s'" % (proc.before.decode('utf-8')))

                    section_transcript += self.strip_newlines(proc.before.decode('utf-8'))
                    start = proc.after.decode('utf-8')
                except Exception:
                    raise
                except pexpect.EOF:
                    self.log_debug("EOF occurred!")
                    raise DexyEOFException()
                except pexpect.TIMEOUT:
                    for c in str(proc.before):
                        print(ord(c), ":", c)
                    msg = "pexpect timeout error. failed at matching prompt within %s seconds. " % timeout
                    msg += "received '%s', tried to match with '%s'" % (proc.before, search_terms)
                    msg += "something may have gone wrong, or you may need to set a longer timeout"
                    self.log_warn(msg)
                    raise UserFeedback(msg)
                except pexpect.ExceptionPexpect as e:
                    raise UserFeedback(str(e))
                except pexpect.EOF as e:
                    raise UserFeedback(str(e))

            if self.setting('strip-regex'):
                section_transcript = re.sub(self.setting('strip-regex'), "", section_transcript)

            yield section_key, section_transcript

        if self.setting('add-new-files'):
            self.add_new_files()

        try:
            proc.close()
        except pexpect.ExceptionPexpect:
            msg = "process %s may not have closed for %s"
            msgargs = (proc.pid, self.key)
            raise UserFeedback(msg % msgargs)

        if proc.exitstatus and self.setting('check-return-code'):
            self.handle_subprocess_proc_return(self.setting('executable'), proc.exitstatus, section_transcript)

    def process(self):
        self.log_debug("about to populate_workspace")
        self.populate_workspace()

        for section_name, section_transcript in self.section_output():
            text = self.strip_trailing_prompts(section_transcript)
            self.log_debug("About to append section %s" % section_name)
            self.output_data[section_name] = text

        self.output_data.save()

try:
    import IPython
    IPython
    IPYTHON_AVAILABLE = True
except ImportError:
    IPYTHON_AVAILABLE = False

class IpythonPexpectReplFilter(PexpectReplFilter):
    """
    Runs python code in the IPython console.
    """
    aliases = ['ipython']
    _settings = {
            'executable' : 'ipython --classic',
            'check-return-code' : False,
            'tags' : ['python', 'repl', 'code'],
            'input-extensions' : [".txt", ".py"],
            'output-extensions' : [".pycon"],
            'version-command' : 'ipython -Version'
            }

    def is_active(klass):
        return IPYTHON_AVAILABLE

class ClojureInteractiveFilter(PexpectReplFilter):
    """
    Runs clojure in REPL.
    """
    aliases = ['cljint']
    _settings = {
            'check-return-code' : False,
            'executable' : 'clojure -r',
            'tags' : ['code', 'clojure', 'repl'],
            'input-extensions' : [".clj", ".txt"],
            'output-extensions' : [".txt"],
            'prompt' : "user=> "
            }

    def lines_for_section(self, input_text):
        input_lines = []
        current_line = []
        in_indented_block = False
        for l in input_text.splitlines():
            if re.match("^\s+", l):
                in_indented_block = True
                current_line.append(l)
            else:
                if len(current_line) > 0:
                    input_lines.append("\n".join(current_line))
                if in_indented_block:
                    # we have reached the end of this indented block
                    in_indented_block = False
                current_line = [l]
        input_lines.append("\n".join(current_line))
        return input_lines

class PythonConsole(PexpectReplFilter):
    """
    Runs python code in python's REPL.
    """

    aliases = ['pycon', 'pyrepl']
    _settings = {
            'check-return-code' : False,
            'tags' : ['repl', 'python', 'code'],
            'executable' : 'python',
            'initial-prompt' : '>>>',
            'input-extensions' : [".txt", ".py"],
            'output-extensions' : ['.pycon'],
            'version-command' : 'python --version',
            'save-vars-to-json-cmd' : """import json
with open("%s-vars.json", "w") as dexy__vars_file:
    dexy__x = {}
    for dexy__k, dexy__v in locals().items():
        try:
            dexy__x[dexy__k] = json.dumps(dexy__v)
        except Exception:
            pass
    json.dump(dexy__x, dexy__vars_file)"""}




from dexy.filter import DexyFilter
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.formatters.latex import LatexFormatter
from pygments.lexers import get_lexer_for_filename
import json

try:
    from lxml import etree
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class XmlSectionFilter(DexyFilter):
    """
    Stores all elements in the input XML document which have any of the
    attributes specified in unique-attributes or qualified-attributes.
    """
    aliases = ["xxml", "xmlsec"]
    _settings = {
            'input-extensions' : [".xml", ".html", ".txt"],
            'pygments' : ("Whether to apply pygments syntax highlighting", True),
            'unique-attributes' : ("Elements to be added if they have this attribute, to be treated as globally unique.", ["id"]),
            'qualified-attributes' : ("Elements to be added if they have this attribute, to be qualified by element type.", ["name"]),
            'data-type' : 'keyvalue',
            'output-extensions' :  [".json", ".sqlite3"]
            }

    def is_active(self):
        return AVAILABLE

    def append_element_attributes_with_key(self, element, element_key):
        source = etree.tostring(element, pretty_print=True).strip()
        inner_html = "\n".join(etree.tostring(child) for child in element.iterchildren())
        self.output_data.append("%s:lineno" % element_key, element.sourceline)
        self.output_data.append("%s:tail" % element_key, element.tail)
        self.output_data.append("%s:text" % element_key, element.text)
        self.output_data.append("%s:tag" % element_key, element.tag)
        self.output_data.append("%s:source" % element_key, source)
        self.output_data.append("%s:inner-html" % element_key, inner_html)

        safe_attrib = {}
        for k, v in element.attrib.items():
            try:
                json.dumps(v)
                safe_attrib[k] = v
            except TypeError:
                pass

        self.output_data.append("%s:attrib" % element_key, json.dumps(safe_attrib))

        if self.setting('pygments'):
            self.output_data.append("%s:html-source" % element_key, highlight(source, self.lexer, self.html_formatter))
            self.output_data.append("%s:latex-source" % element_key, highlight(source, self.lexer, self.latex_formatter))

    def process(self):
        assert self.output_data.state == 'ready'

        if self.setting('pygments'):
            self.lexer = get_lexer_for_filename(self.input_data.storage.data_file())
            self.html_formatter = HtmlFormatter(lineanchors=self.output_data.web_safe_document_key())
            self.latex_formatter = LatexFormatter()

        if self.input_data.ext in ('.xml', '.txt'):
            parser = etree.XMLParser()
        elif self.input_data.ext == '.html':
            parser = etree.HTMLParser()
        else:
            raise Exception("Unsupported extension %s" % self.input_data.ext)

        tree = etree.parse(self.input_data.storage.data_file(), parser)

        for element in tree.iter("*"):
            element_keys = []
           
            for attribute_name in self.setting('unique-attributes'):
                if attribute_name in element.attrib:
                    element_keys.append(element.attrib[attribute_name])
            for attribute_name in self.setting('qualified-attributes'):
                if attribute_name in element.attrib:
                    element_keys.append(element.attrib[attribute_name])
                    element_keys.append("%s:%s" % (element.tag, element.attrib[attribute_name]))

            for element_key in element_keys:
                self.append_element_attributes_with_key(element, element_key)

        self.output_data.save()



from dexy.filters.process import SubprocessExtToFormatFilter
from dexy.filters.process import SubprocessFilter
from dexy.filters.process import SubprocessFormatFlagFilter
from dexy.filters.process import SubprocessInputFilter
from dexy.filters.process import SubprocessStdoutFilter
from dexy.utils import file_exists
import dexy.exceptions
import json
import os
import shutil

class Kramdown(SubprocessStdoutFilter):
    """
    Runs the kramdown markdown converter.

    http://kramdown.gettalong.org/
    """
    aliases = ['kramdown']
    _settings = {
            'added-in-version' : "1.0.5",
            'class' : 'SubprocessExtToFormatFilter',
            'executable' : 'kramdown',
            'version-command' : 'kramdown -version',
            'output' : True,
            'format-specifier' : '-o ',
            'ext-to-format' : {'.tex' : 'latex', '.html' : 'html', '.md' : 'kramdown', '.txt' : 'kramdown'},
            'input-extensions' : [".*"],
            'output-extensions' : [".html", ".tex", ".txt", ".md"],
            'require-output' : True,
            'command-string' : '%(prog)s %(format)s %(args)s "%(script_file)s"'
            }

    def command_string_args(self):
        args = self.default_command_string_args()

        fmt_specifier = self.setting('format-specifier')
        if fmt_specifier and (fmt_specifier in args['args']):
            # Already have specified the format manually.
            fmt = ''
        else:
            fmt_setting = self.setting('ext-to-format')[self.ext]
            if fmt_setting:
                fmt = "%s%s" % (fmt_specifier, fmt_setting)
            else:
                fmt = ''

        args['format'] = fmt
        return args

class Redcarpet(SubprocessStdoutFilter):
    """
    Converts github-flavored markdown to HTML using redcarpet.
    """
    aliases = ['redcarpet', 'ghmd']
    _settings = {
            'added-in-version' : '1.0.1',
            'executable' : 'redcarpet',
            'input-extensions' : [".md", ".txt"],
            'output-extensions' : [".html"],
            # parse options
            'parse-autolink' : ("autolink option from redcarpet", True),
            'parse-disabled-indented-code-blocks' : ("disabled-indented-code-blocks option from redcarpet", False),
            'parse-fenced-code-blocks' : ("fenced-code-blocks option from redcarpet", True),
            'parse-highlight' : ("highlight option from redcarpet", True),
            'parse-lax-spacing' : ("lax-spacing option from redcarpet", True),
            'parse-no-intra-emphasis' : ("no-intra-emphasis option from redcarpet", True),
            'parse-quotes' : ("quotes option from redcarpet", True),
            'parse-space-after-headers' : ("space-after-headers option from redcarpet", True),
            'parse-strikethrough' : ("strikethrough option from redcarpet", True),
            'parse-subscript' : ("subscript option from redcarpet", True),
            'parse-superscript' : ("superscript option from redcarpet", True),
            'parse-tables' : ("tables option from redcarpet", True),
            'parse-underline' : ("underline option from redcarpet", True),
            # render options
            'render-filter-html' : ("filter-html option from redcarpet", False),
            'render-no-images' : ("no-images option from redcarpet", False),
            'render-no-links' : ("no-links option from redcarpet", False),
            'render-no-styles' : ("no-styles option from redcarpet", False),
            'render-safe-links-only' : ("safe-links-only option from redcarpet", False),
            'render-with-toc-data' : ("with-toc-data option from redcarpet", False),
            'render-hard-wrap' : ("hard-wrap option from redcarpet", False),
            'render-prettify' : ("prettify option from redcarpet", False),
            'render-xhtml' : ("xhtml option from redcarpet", False),
            # other options
            'pygments' : ("Pygments syntax highlighting (requires ananelson/redcarpet fork).", False)
            }
  
    def command_string(self):
        args = self.command_string_args()

        args['parse_args'] = " ".join("--%s" % name for name in args if name.startswith("parse-") and args[name])
        args['render_args'] = " ".join("--%s" % name for name in args if name.startswith("render-") and args[name])

        other_args = ['pygments']
        args['other_args'] = " ".join("--%s" % name for name in other_args if args[name])

        return "%(prog)s %(parse_args)s %(render_args)s %(other_args)s %(script_file)s" % args

class TidyCheck(SubprocessFilter):
    """
    Runs `tidy` to check for valid HTML.

    This filter does not alter valid HTML. It raises an Exception if invalid
    HTML is found.
    """
    aliases = ['tidycheck']
    _settings = {
        'examples' : ['tidy'],
        'tags' : ['html'],
        'executable' : 'tidy',
        'command-string' : '%(prog)s -errors -quiet "%(script_file)s"',
        'input-extensions' : ['.html'],
        'output-extensions' : ['.txt']
        }

    def process(self):
        command = self.command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)

        # If we get here, just return original input.
        self.output_data.copy_from_file(self.input_data.storage.data_file())

class PdfToCairo(SubprocessFormatFlagFilter):
    """
    Runs `pdftocairo` from the poppler library.

    Converts PDF input to various output formats inclusing SVG.
    """
    aliases = ['pdftocairo', 'pdf2cairo', 'pdf2svg', 'pdftosvg']
    _settings = {
        'command-string': '%(prog)s %(format)s %(args)s "%(script_file)s" "%(output_file)s"',
        'executable': 'pdftocairo',
        'tags' : ['pdf', 'image'],
        'input-extensions' : ['.pdf'],
        'output-extensions' : ['.svg', '.png', '.jpg', '.ps', '.eps', '.pdf'],
        'ext-to-format' : {
            '.png' : '-png',
            '.jpg' : '-jpeg',
            '.ps' : '-ps',
            '.eps' : '-eps',
            '.pdf' : '-pdf',
            '.svg' : '-svg'
            }
        }

    def process(self):
        command = self.command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)

        if not self.output_data.is_cached():
            # Find the first page
            for pagenum in ('1', '01', '001', '0001',):
                basename = os.path.join(self.workspace(), self.output_data.name)
                first_page_file = "%s-%s.png" % (basename, pagenum)
                if file_exists(first_page_file):
                    print("Copy from '%s'" % first_page_file)
                    self.output_data.copy_from_file(first_page_file)
                    break

        assert self.output_data.is_cached()

        if self.setting('add-new-files'):
            self.log_debug("adding new files found in %s for %s" % (self.workspace(), self.key))
            self.add_new_files()

class Pdf2ImgSubprocessFilter(SubprocessExtToFormatFilter):
    """
    Runs ghostscript to convert PDF files to images.

    An image file can only hold a single page of PDF, so this defaults to
    returning page 1. The `page` setting can be used to specify other pages.
    """
    aliases = ['pdf2img', 'pdftoimg', 'pdf2png']
    _settings = {
            'res' : ("Resolution of image.", 300),
            'page' : ("Which page of the PDF to return as an image", 1),
            'executable' : 'gs',
            'version-command' : 'gs --version',
            'tags' : ['pdf', 'gs'],
            'input-extensions' : ['.pdf'],
            'output-extensions' : ['.png'],
            'ext-to-format' : {
                '.png' : 'png16m',
                '.jpg' : 'jpeg'
                },
            'format-specifier' : '-sDEVICE=',
            'command-string' : '%(prog)s -dSAFER -dNOPAUSE -dBATCH %(format)s -r%(res)s -sOutputFile="%%d-%(output_file)s" "%(script_file)s"'
            }

    def process(self):
        self.populate_workspace()

        command = self.command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)

        page = self.setting('page')
        page_file = "%s-%s" % (page, self.output_data.basename())

        wd = self.parent_work_dir()
        page_path = os.path.join(wd, page_file)
        shutil.copyfile(page_path, self.output_filepath())

class RIntBatchSectionsFilter(SubprocessFilter):
    """
    Experimental filter to run R in sections without using pexpect.
    """
    aliases = ['rintmock']

    _settings = {
            'add-new-files' : True,
            'executable' : 'R CMD BATCH --quiet --no-timing',
            'tags' : ['rstats', 'repl', 'stats'],
            'input-extensions' : ['.txt', '.r', '.R'],
            'output-extensions' : [".Rout", '.txt'],
            'version-command' : "R --version",
            'write-stderr-to-stdout' : False,
            'data-type' : 'sectioned',
            'command-string' : """%(prog)s %(args)s "%(script_file)s" %(scriptargs)s "%(output_file)s" """
            }

    def command_string(self, section_name, section_text, wd):
        br = self.input_data.baserootname()

        args = self.default_command_string_args()
        args['script_file'] = "%s-%s%s" % (br, section_name, self.input_data.ext)
        args['output_file'] = "%s-%s-out%s" % (br, section_name, self.output_data.ext)

        work_filepath = os.path.join(wd, args['script_file'])

        with open(work_filepath, "w") as f:
            f.write(str(section_text))

        command = self.setting('command-string') %  args
        return command, args['output_file']

    def process(self):
        self.populate_workspace()
        wd = self.parent_work_dir()

        for section_name, section_text in self.input_data.items():
            command, outfile = self.command_string(section_name, section_text, wd)
            proc, stdout = self.run_command(command, self.setup_env())
            self.handle_subprocess_proc_return(command, proc.returncode, stdout)

            with open(os.path.join(wd, outfile), "r") as f:
                self.output_data[section_name] = f.read()

        if self.setting('add-new-files'):
            self.add_new_files()

        self.output_data.save()

class EmbedFonts(SubprocessFilter):
    """
    Runs ghostscript ps2pdf with prepress settings.

    Allegedly this helps embed fonts and makes documents friendly for printing.
    """
    aliases = ['embedfonts', 'prepress']
    _settings = {
            'input-extensions' : [".pdf"],
            'output-extensions' : [".pdf"],
            'executable' : 'ps2pdf',
            'tags' : ['pdf'],
            }

    def preprocess_command_string(self):
        pf = self.work_input_filename()
        af = self.work_output_filename()
        return "%s -dPDFSETTINGS=/prepress %s %s" % (self.setting('executable'), pf, af)

    def pdffonts_command_string(self):
        return "%s %s" % ("pdffonts", self.result().name)

    def process(self):
        env = self.setup_env()

        command = self.preprocess_command_string()
        proc, stdout = self.run_command(command, env)
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)

        command = self.pdffonts_command_string()
        proc, stdout = self.run_command(command, env)
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)

        self.copy_canonical_file()

class AbcFilter(SubprocessFormatFlagFilter):
    """
    Runs `abcm2ps` on .abc music files.
    """
    aliases = ['abc']
    _settings = {
            'command-string' : '%(prog)s %(args)s %(format)s -O %(output_file)s %(script_file)s',
            'add-new-files' : False,
            'output' : True,
            'tags' : ['music'],
            'examples' : ['abc'],
            'executable' : 'abcm2ps',
            'input-extensions' : ['.abc'],
            'output-extensions': ['.svg', '.html', '.xhtml', '.eps'],
            'ext-to-format': {
                '.eps' : '-E',
                '.svg' : '-g',
                '.svg1' : '-v', # dummy entry so we know -v is a format flag
                '.html' : '-X',
                '.xhtml' : '-X'
                }
            }

    def process(self):
        command = self.command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)

        if self.ext in ('.svg', '.eps'):
            # Fix for abcm2ps adding 001 to file name.
            nameparts = os.path.splitext(self.output_data.name)
            output_filename = "%s001%s" % (nameparts[0], nameparts[1])
            output_filepath = os.path.join(self.workspace(), output_filename)
            self.output_data.copy_from_file(output_filepath)
        else:
            self.copy_canonical_file()

        if self.setting('add-new-files'):
            self.add_new_files()

class AbcMultipleFormatsFilter(SubprocessFilter):
    """
    Runs `abcm2ps` on .abc music files, generating all output formats.
    """
    aliases = ['abcm']
    _settings = {
            'input-extensions' : ['.abc'],
            'output-extensions' : ['.json'],
            'executable' : 'abcm2ps',
            'tags' : ['music'],
            'add-new-files' : False
            }

    def command_string(self, ext):
        clargs = self.command_line_args() or ''

        if any(x in clargs for x in ['-E', '-g', '-v', '-X']):
            raise dexy.exceptions.UserFeedback("Please do not pass any output format flags!")

        if ext in ('.eps'):
            output_flag = '-E'
        elif ext in ('.svg'):
            output_flag = '-g'
        elif ext in ('.html', '.xhtml'):
            output_flag = '-X'
        else:
            raise dexy.exceptions.InternalDexyProblem("bad ext '%s'" % ext)

        args = {
            'prog' : self.setting('executable'),
            'args' : clargs,
            'output_flag' : output_flag,
            'script_file' : self.work_input_filename(),
            'output_file' : self.output_workfile(ext)
        }
        return "%(prog)s %(args)s %(output_flag)s -O %(output_file)s %(script_file)s" % args

    def output_workfile(self, ext):
        return "%s%s" % (self.output_data.baserootname(), ext)

    def process(self):
        output = {}

        wd = self.parent_work_dir()

        for ext in ('.eps', '.svg', '.html', '.xhtml'):
            command = self.command_string(ext)
            proc, stdout = self.run_command(command, self.setup_env())
            self.handle_subprocess_proc_return(command, proc.returncode, stdout)

            if ext in ('.svg', '.eps'):
                # Fix for abcm2ps adding 001 to file name.
                nameparts = os.path.splitext(self.output_workfile(ext))
                output_filename = "%s001%s" % (nameparts[0], nameparts[1])
                output_filepath = os.path.join(wd, output_filename)
            else:
                output_filename = self.output_workfile(ext)
                output_filepath = os.path.join(wd, output_filename)

            with open(output_filepath, "r") as f:
                output[ext] = f.read()

        self.output_data.set_data(json.dumps(output))

class ManPage(SubprocessStdoutFilter):
    """
    Read command names from a file and fetch man pages for each.

    Returns a JSON dict whose keys are the program names and values are man
    pages.
    """
    aliases = ['man']

    _settings = {
            'executable' : 'man',
            'tags' : ['utils'],
            'version-command' : 'man --version',
            'input-extensions' : [".txt"],
            'output-extensions' : [".json"]
    }

    def command_string(self, prog_name):
        # Use bash rather than the default of sh (dash) so we can set pipefail.
        return "bash -c \"set -e; set -o pipefail; man %s | col -b | strings\"" % (prog_name)

    def process(self):
        man_info = {}
        for prog_name in str(self.input_data).split():
            command = self.command_string(prog_name)
            proc, stdout = self.run_command(command, self.setup_env())
            self.handle_subprocess_proc_return(command, proc.returncode, stdout)
            man_info[prog_name] = stdout

        self.output_data.set_data(json.dumps(man_info))

class ApplySed(SubprocessInputFilter):
    """
    Runs `sed` on the input file.

    Expects a sed script to be a dependency.
    """
    aliases = ['used']
    _settings = {
            'executable' : 'sed',
            'tags' : ['utils'],
            'data-type' : 'generic',
            }

    def process(self):
        for doc in self.doc.walk_input_docs():
            if doc.output_data().ext == ".sed":
                command = "%s -f %s" % (self.setting('executable'), doc.name)

        if not command:
            raise dexy.exceptions.UserFeedback("A .sed file must be passed as an input to %s" % self.key)

        proc, stdout = self.run_command(command, self.setup_env(), str(self.input_data))
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)
        self.output_data.set_data(stdout.rstrip("\n"))

class Sed(SubprocessInputFilter):
    """
    Runs a sed script.

    Any dependencies are assumed to be text files and they have the sed script
    applied to them.
    """
    aliases = ['sed']
    _settings = {
            'executable' : 'sed',
            'tags' : ['utils'],
            'input-extensions' : ['.sed'],
            'output-extensions' : ['.sed', '.txt'],
            }

    def command_string(self):
        return "%s -f %s" % (self.setting('executable'), self.work_input_filename())

class Taverna(SubprocessStdoutFilter):
    """
    Runs workflows in Taverna via command line tool.
    """
    aliases = ['taverna']
    _settings = {
            'executable' : 'taverna',
            'tags' : ['repro', 'workflow'],
            'add-new-files' : True,
            'input-extensions' : ['.t2flow'],
            'output-extensions' : ['.txt'],
            'taverna-home' : ("Location of taverna home directory.", "$TAVERNA_HOME"),
            'x-max' : ("Java -Xmx setting", '300m'),
            'x-perm-max' : ("Java -XX:MaxPermSize setting", '140m'),
            }

    def command_string(self):
        assert self.setting('taverna-home')

        return """java -Xmx%(x-max)s -XX:MaxPermSize=%(x-perm-max)s \\
                -Draven.profile=file://%(taverna-home)s/conf/current-profile.xml \\
                -Dtaverna.startup=%(taverna-home)s \\
                -Djava.system.class.loader=net.sf.taverna.raven.prelauncher.BootstrapClassLoader \\
                -Draven.launcher.app.main=net.sf.taverna.t2.commandline.CommandLineLauncher \\
                -Draven.launcher.show_splashscreen=false \\
                -Djava.awt.headless=true \\
                -jar "%(taverna-home)s/lib/"prelauncher-*.jar \\
                %(script_file)s""" % self.command_string_args()



from dexy.utils import file_exists
import dexy.filter
import json
import os

class ApiFilter(dexy.filter.DexyFilter):
    """
    Base class for filters which post content to a remote API.
    
    This class provides standard formats and locations for storing
    configuration and authentication information.

    Need to read config for the API in general, such as the base URL and an API
    key for authentication.

    Also need to read config for the particular task/document being uploaded.
    This could be stored in the .dexy config entry, but this has two drawbacks.
    First, it makes it difficult to bulk define documents according to a
    pattern. Secondly, it makes it very difficult for dexy to modify this
    configuration to add additional information, such as the returned id for a
    newly created document.

    So, it is preferable to define a local file (defined by a relative path to
    the document in question) which can be overridden per-document in which we
    just store the API-related config and which can be modified by dexy without
    concern about identifying the entry in a .dexy file or accidentally
    overwriting some unrelated information.
    """
    aliases = ['apis']

    _settings = {
            # Files to hold collections of API keys
            'master-api-key-file' : ("Master API key file for user.", "~/.dexyapis"),
            'project-api-key-file' : ("API key file for project.", ".dexyapis"),

            # Parameters to be stored in collection files
            'api-username' : ("The username to sign into the API with.", None),
            'api-password' : ("The password to sign into the API with.", None),
            'api-url' : ("The url of the API endpoint.", None),

            # Files to hold info about a single document
            'document-api-config-file' : ('Filename to store config for a file (can only have 1 per directory, dexy looks for suffix format first.', None),
            'document-api-config-postfix' : ('Suffix to attach to content filename to indicate this is the config for that file.', '-config.json'),

            'api-key-name' : ("The name of this API", None),
            }

    def api_key_locations(self):
        return [self.setting('project-api-key-file'), self.setting('master-api-key-file')]

    def docmd_create_keyfile(self):
        return self.create_keyfile('master-api-key-file')

    def create_keyfile(self, keyfilekey):
        """
        Creates a key file.
        """
        key_filename = os.path.expanduser(self.setting(keyfilekey))
        if file_exists(key_filename):
            raise dexy.exceptions.UserFeedback("File %s already exists!" % key_filename)

        keyfile_content = {}
        for filter_instance in dexy.filter.Filter:
            if isinstance(filter_instance, ApiFilter) and not filter_instance.__class__ == ApiFilter:
                api_key_name = filter_instance.setting('api-key-name')
                # We want to create a keyfile for this filter instance.
                keyfile_content[api_key_name] = {}

                # Get all the entries we want in the keyfile.
                for k, v in filter_instance.setting_values().items():
                    if k.startswith("api_"):
                        keyfile_content[api_key_name][k.replace("api_", "")] = "TODO"

        with open(key_filename, "w") as f:
            json.dump(keyfile_content, f, sort_keys = True, indent=4)

    def document_config_file(self):
        postfix_config_filename = "%s%s" % (os.path.splitext(self.output_data.name)[0], self.setting('document-api-config-postfix'))
        if file_exists(postfix_config_filename):
            return postfix_config_filename
        else:
            return os.path.join(self.output_data.parent_dir(), self.setting('document-api-config-file'))

    def read_document_config(self):
        document_config = self.document_config_file()
        if file_exists(document_config):
            with open(document_config, "r") as f:
                return json.load(f)
        else:
            msg = "Filter %s needs a file %s, couldn't find it."
            raise dexy.exceptions.UserFeedback(msg % (self.alias, document_config))

    def save_document_config(self, config):
        document_config = self.document_config_file()
        with open(document_config, "w") as f:
            json.dump(config, f, sort_keys=True, indent=4)

    def read_param(self, param_name):
        param_value = None

        for filename in self.api_key_locations():
            if "~" in filename:
                filename = os.path.expanduser(filename)

            if file_exists(filename):
                with open(filename, "r") as f:
                    params = json.load(f)
                    if self.setting('api-key-name') in params:
                        param_value = params[self.setting('api-key-name')].get(param_name)

            if param_value:
                break

        if param_value and isinstance(param_value, str) and param_value.startswith("$"):
            # need to get value of bash variable
            param_value_from_env = os.getenv(param_value.lstrip("$"))
            if not param_value_from_env:
                raise KeyError("Bash variable %s not defined in this environment!" % param_value)
            param_value = param_value_from_env

        if param_value:
            return param_value
        else:
            msg = "Could not find %s for %s in: %s" % (param_name, self.setting('api-key-name'), ", ".join(self.api_key_locations()))
            raise KeyError(msg)



from tests.utils import wrap
from mock import patch
import dexy.filter
import os

@patch('os.path.expanduser')
def test_docmd_create_keyfile(mod):
    mod.return_value = '.dexyapis'
    with wrap():
        assert not os.path.exists(".dexyapis")
        dexy.filter.Filter.create_instance("apis").docmd_create_keyfile()
        assert os.path.exists(".dexyapis")



import uuid
import pickle
import os
import dexy.data

class Batch(object):
    def __init__(self, wrapper):
        self.wrapper = wrapper
        self.docs = {}
        self.doc_keys = {}
        self.filters_used = []
        self.uuid = str(uuid.uuid4())
        self.start_time = None
        self.end_time = None

    def __repr__(self):
        return "Batch(%s)" % self.uuid

    def __iter__(self):
        for doc_key in self.docs:
            if self.docs[doc_key]['state'] in ('uncached',):
                continue
            yield self.output_data(doc_key)

    def add_doc(self, doc):
        """
        Adds a new doc to the batch of docs.
        """
        if hasattr(doc, 'batch_info'):
            doc_key = doc.key_with_class()
            storage_key = doc.output_data().storage_key
            self.doc_keys[storage_key] = doc_key
            self.update_doc_info(doc)
            self.filters_used.extend(doc.filter_aliases)

    def update_doc_info(self, doc):
        self.docs[doc.key_with_class()] = doc.batch_info()

    def output_data(self, doc_key):
        return self.data(doc_key, 'output')

    def input_data(self, doc_key):
        return self.data(doc_key, 'input')

    def doc_info(self, doc_key):
        return self.docs[doc_key]
   
    def doc_key(self, storage_key):
        return self.doc_keys[storage_key]

    def data_for_storage_key(self, storage_key, input_or_output='output'):
        """
        Retrieves a data object given the storage key (based on the the
        md5sum), rather than the canonical doc key based on the doc name.
        """
        doc_key = self.doc_key(storage_key)
        return self.data(doc_key, input_or_output)

    def data(self, doc_key, input_or_output='output'):
        """
        Retrieves a data object given the doc key.
        """
        doc_info = self.doc_info(doc_key)["%s-data" % input_or_output]
        args = list(doc_info)
        args.append(self.wrapper)
        data = dexy.data.Data.create_instance(*args)
        data.setup_storage()
        if hasattr(data.storage, 'connect'):
            data.storage.connect()
        return data

    def elapsed(self):
        if self.end_time and self.start_time:
            return self.end_time - self.start_time
        else:
            return 0

    def filename(self):
        return "%s.pickle" % self.uuid

    def filepath(self):
        return os.path.join(self.batch_dir(), self.filename())

    def most_recent_filename(self):
        return os.path.join(self.batch_dir(), 'most-recent-batch.txt')

    def batch_dir(self):
        return os.path.join(self.wrapper.artifacts_dir, 'batches')

    def to_dict(self):
        attr_names = ['docs', 'doc_keys', 'filters_used', 'uuid']
        return dict((k, getattr(self, k),) for k in attr_names)

    def save_to_file(self):
        try:
            os.makedirs(self.batch_dir())
        except OSError:
            pass

        with open(self.filepath(), 'wb') as f:
            pickle.dump(self.to_dict(), f)

        with open(self.most_recent_filename(), 'w') as f:
            f.write(self.uuid)

    def load_from_file(self):
        with open(self.filepath(), 'rb') as f:
            d = pickle.load(f)
            for k, v in d.items():
                setattr(self, k, v)

    @classmethod
    def load_most_recent(klass, wrapper):
        """
        Retuns a batch instance representing the most recent batch as indicated
        by the UUID stored in most-recent-batch.txt.
        """
        batch = Batch(wrapper)
        try:
            with open(batch.most_recent_filename(), 'r') as f:
                most_recent_uuid = f.read()

            batch.uuid = most_recent_uuid
            batch.load_from_file()
            return batch
        except IOError:
            pass



from dexy.exceptions import InternalDexyProblem
import dexy.plugin
import dexy.storage
import dexy.utils
import dexy.wrapper
import inflection
import os
import posixpath
import shutil
import urllib

class Data(dexy.plugin.Plugin, metaclass=dexy.plugin.PluginMeta):
    """
    Base class for types of Data.
    """
    _settings = {
            'shortcut' : ("A shortcut to refer to a file.", None),
            'storage-type' : ("Type of storage to use.", 'generic'),
            'canonical-output' : ("Whether this data type is canonical output.", None),
            'canonical-name' : ("The default name.", None),
            'output-name' : ("A custom name which overrides default name.", None),
            'title' : ("A custom title.", None),
            }

    state_transitions = (
            (None, 'new'),
            ('new', 'ready'),
            ('ready', 'ready')
            )

    def add_to_lookup_nodes(self):
        if self.setting('canonical-output'):
            self.wrapper.add_data_to_lookup_nodes(self.key, self)
            self.wrapper.add_data_to_lookup_nodes(self.output_name(), self)
            self.wrapper.add_data_to_lookup_nodes(self.title(), self)

    def add_to_lookup_sections(self):
        if self.setting('canonical-output'):
            for section_name in self.keys():
                if not section_name == '1':
                    self.wrapper.add_data_to_lookup_sections(section_name, self)

    def __init__(self, key, ext, storage_key, settings, wrapper):
        self.key = key
        self.ext = ext
        self.storage_key = storage_key

        self.wrapper = wrapper
        self.initialize_settings(**settings)
        self.update_settings(settings)

        self._data = None
        self.state = None
        self.name = self.setting('canonical-name')
        if not self.name:
            msg = "Document must provide canonical-name setting to data."
            raise InternalDexyProblem(msg)
        elif self.name.startswith("("):
            raise Exception()

        self.transition('new')
        
    def transition(self, new_state):
        """
        Transition between states in a state machine.
        """
        dexy.utils.transition(self, new_state)

    def args_to_data_init(self):
        """
        Returns tuple of attributes to pass to create_instance.
        """
        return (self.alias, self.key, self.ext, self.storage_key, self.setting_values())

    def setup(self):
        self.setup_storage()
        self.transition('ready')

    def setup_storage(self):
        storage_type = self.storage_class_alias(self.ext)
        instanceargs = (self.storage_key, self.ext, self.wrapper,)
        self.storage = dexy.storage.Storage.create_instance(storage_type, *instanceargs)

        self.storage.assert_location_is_in_project_dir(self.name)

        if self.output_name():
            self.storage.assert_location_is_in_project_dir(self.output_name())

        self.storage.setup()

    def storage_class_alias(self, file_ext):
        return self.setting('storage-type')

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self.key)

    def __lt__(self, other):
        """
        Sort data obejects by their output name.
        """
        return self.output_name() < other.output_name()

    def __getattr__(self, item):
        """
        Make setting values available as attributes.
        """
        try:
            return self._instance_settings[item]
        except KeyError:
            raise AttributeError(item)

    def data(self):
        if (not self._data) or self._data == [{}]:
            self.load_data()
        return self._data

    def load_data(self, this=None):
        try:
            self._data = self.storage.read_data()
        except IOError:
            msg = "no data in file '%s' for %s (wrapper state '%s', data state '%s')"
            msgargs = (self.storage.data_file(), self.key,
                    self.wrapper.state, self.state)
            raise dexy.exceptions.InternalDexyProblem(msg % msgargs)

    def clear_data(self):
        self._data = None

    def clear_cache(self):
        self._size = None
        try:
            os.remove(self.storage.data_file())
        except os.error as e:
            self.wrapper.log.warn(str(e))

    def copy_from_file(self, filename):
        shutil.copyfile(filename, self.storage.data_file())

    def output_to_file(self, filepath):
        """
        Write canonical output to a file. Parent directory must exist already.
        """
        if not self.storage.copy_file(filepath):
            self.storage.write_data(self.data(), filepath)

    def has_data(self):
        has_loaded_data = (self._data) and (self._data != [{}])
        return has_loaded_data or self.is_cached()

    def set_data(self, data):
        """
        Shortcut to set and save data.
        """
        self._data = data
        self.save()

    def is_cached(self, this=None):
        if this is None:
            this = (self.wrapper.state in ('walked', 'running'))
        return self.storage.data_file_exists(this)

    # Filename-related Attributes

    def parent_dir(self):
        """
        The name of the directory containing the document.
        """
        return posixpath.dirname(self.name)

    def parent_output_dir(self):
        """
        The name of the directory containing the document based on final output
        name, which may be specified in a different directory.
        """
        return posixpath.dirname(self.output_name())

    def long_name(self):
        """
        A unique, but less canonical, name for the document.
        """
        if "|" in self.key:
            return "%s%s" % (self.key.replace("|", "-"), self.ext)
        else:
            return self.setting('canonical-name')

    def rootname(self):
        """
        Returns the file name, including path, without extension.
        """
        return os.path.splitext(self.name)[0]

    def basename(self):
        """
        Returns the local file name without path.
        """
        return posixpath.basename(self.name)

    def baserootname(self):
        """
        Returns local file name without extension or path.
        """
        return posixpath.splitext(self.basename())[0]

    def web_safe_document_key(self):
        """
        Returns document key with slashes replaced by double hypheens.
        """
        return self.long_name().replace("/", "--")

    def title(self):
        """
        Canonical title of document.

        Tries to guess from document name if `title` setting not provided.
        """
        if self.setting('title'):
            return self.setting('title')

        if self.is_index_page():
            subdir = posixpath.split(posixpath.dirname(self.name))[-1]
            if subdir == "/":
                return "Home"
            elif subdir:
                return inflection.titleize(subdir)
            else:
                return inflection.titleize(self.baserootname())
        else:
            return inflection.titleize(self.baserootname())

    def relative_path_to(self, relative_to):
        """
        Returns a relative path from this document to the passed other
        document.
        """
        return posixpath.relpath(relative_to, self.output_parent_dir())

    def strip(self):
        """
        Returns contents stripped of leading and trailing whitespace.
        """
        return str(self).strip()
    
    def splitlines(self, arg=None):
        """
        Returns a list of lines split at newlines or custom split.
        """
        return str(self).splitlines(arg)

    def url_quoted_name(self):
        """
        Applies urllib's quote method to name.
        """
        return urllib.quote(self.name)

    def output_name(self):
        """
        Canonical name to output to, relative to output root. Returns None if
        artifact not in output_root.
        """
        output_root = self.wrapper.output_root

        def relativize(path):
            if output_root == ".":
                return path
            elif os.path.abspath(output_root) in os.path.abspath(path):
                return os.path.relpath(path, output_root)

        output_name = self.setting('output-name')
        if output_name:
            return relativize(output_name)
        else:
            return relativize(self.name)

    def output_parent_dir(self):
        """
        Canonical output directory, taking into account custom outputroot and document name.
        """
        return os.path.dirname(self.output_name())
        
    def filesize(self, this=None):
        """
        Returns size of file stored on disk.
        """
        if this is None:
            this = (self.wrapper.state in ('walked', 'running'))
        return self.storage.data_file_size(this)

    def is_canonical_output(self):
        """
        Used by reports to determine if document should be written to output/
        directory.
        """
        return self.setting('canonical-output')

    def is_index_page(self):
        """
        Is this a website index page, i.e. named `index.html`.
        """
        return self.output_name() and self.output_name().endswith("index.html")

    def websafe_key(self):
        """
        Returns a web-friendly version of the key.
        """
        return self.key

    # Deprecated methods

    def as_text(self):
        """
        DEPRECATED. Instead call unicode.
        """
        return str(self)

class Generic(Data):
    """
    Data type representing generic binary or text-based data in a single blob.
    """
    aliases = ['generic']

    def save(self):
        if isinstance(self._data, str):
            self.storage.write_data(self._data.encode("utf-8"))
        else:
            if self._data == None:
                msg = "No data found for '%s', did you reference a file that doesn't exist?"
                raise dexy.exceptions.UserFeedback(msg % self.key)
            self.storage.write_data(self._data)

    def __str__(self):
        if isinstance(self.data(), str):
            return self.data()
        elif not self.data():
            return str(None)
        else:
            return self.wrapper.decode_encoded(self.data())

    def iteritems(self):
        """
        Iterable list of sections in document.
        """
        yield ('1', self.data())

    def items(self):
        """
        List of sections in document.
        """
        return [('1', self.data(),)]

    def keys(self):
        """
        List of keys (section names) in document.
        """
        return ['1']

    def __getitem__(self, key):
        if key == '1':
            return self.data()
        else:
            try:
                return self.data()[key]
            except TypeError:
                if self.ext == '.json':
                    return self.from_json()[key]
                elif self.ext == '.yaml':
                    return self.from_yaml()[key]
                else:
                    raise

    def from_json(self):
        """
        Attempts to load data using a JSON parser, returning whatever objects
        are defined in the JSON.
        """
        if self._data and isinstance(self._data, str):
            return dexy.utils.parse_json(self._data)
        elif self._data and not isinstance(self._data, str):
            raise Exception(self._data.__class__.__name__)
        else:
            with open(self.storage.data_file(), "r") as f:
                return dexy.utils.parse_json_from_file(f)

    def from_yaml(self):
        """
        Attempts to load data using a YAML parser, returning whatever objects
        are defined in the YAML.
        """
        if self._data and isinstance(self._data, str):
            return dexy.utils.parse_yaml(self._data)
        elif self._data and not isinstance(self._data, str):
            raise Exception(self._data.__class__.__name__)
        else:
            with open(self.storage.data_file(), "r") as f:
                return dexy.utils.parse_yaml(f.read())

    def json_as_dict(self):
        """
        DEPRECATED. Instead call from_json
        """
        return self.from_json()

class SectionValue(object):
    def __init__(self, data, parent, parentindex):
        assert isinstance(data, dict)
        self.data = data
        self.parent = parent
        self.parentindex = parentindex

    def __str__(self):
        return self.data['contents'] or ''

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.parent.data()[self.parentindex+1][key] = value

    def splitlines(self):
        return str(self).splitlines()

class Sectioned(Data):
    """
    A document with named, ordered sections.

    Sections can also contain arbitrary metadata.
    """
    aliases = ['sectioned']

    _settings = {
            'storage-type' : 'jsonsectioned'
            }

    def setup(self):
        self.setup_storage()
        self._data = [{}]
        self.transition('ready')

    def save(self):
        try:
            self.storage.write_data(self._data)
        except Exception as e:
            print(self._data)
            msg = "Problem saving '%s': %s" % (self.key, str(e))
            raise dexy.exceptions.InternalDexyProblem(msg)

    def __str__(self):
        return "\n".join(str(v) for v in list(self.values()) if str(v))

    def __len__(self):
        """
        The number of sections.
        """
        return len(self.data())-1

    def __setitem__(self, key, value):
        keyindex = self.keyindex(key)
        if keyindex >= 0:
            # Existing section.
            assert self._data[keyindex+1]['name'] == key
            self._data[keyindex+1]['contents'] = value
        else:
            # New section.
            section_dict = {"name" : key, "contents" : value}
            self._data.append(section_dict)

    def __delitem__(self, key):
        index = self.keyindex(key)
        self.data().pop(index+1)

    def keys(self):
        return [a['name'] for a in self.data()[1:]]

    def values(self):
        return [SectionValue(a, self, i) for i, a in enumerate(self.data()[1:])]

    def output_to_file(self, filepath):
        """
        Write canonical (not structured) output to a file.
        """
        with open(filepath, "wb") as f:
            f.write(str(self).encode("utf-8"))

    def keyindex(self, key):
        if self._data == [{}]:
            return -1

        try:
            return list(self.keys()).index(key)
        except ValueError:
            return -1

    def value(self, key, throwException=True):
        index = self.keyindex(key)
        if index > -1:
            return list(self.values())[index]
        else:
            try:
                return self.data()[0][key]
            except KeyError:
                msg = "No value for %s available in sections or metadata."
                msgargs = (key)
                if throwException:
                    raise dexy.exceptions.UserFeedback(msg % msgargs)

    def __getitem__(self, key):
        try:
            return self.data()[key+1]
        except (TypeError, KeyError):
            return self.value(key)

    def items(self):
        """
        Iterable list of sections in document.
        """
        keys = self.keys()
        values = self.values()
        for i in range(len(keys)):
            yield (keys[i], values[i])

class KeyValue(Data):
    """
    Data class for key-value data.
    """
    aliases  = ['keyvalue']
    _settings = {
            'storage-type' : 'sqlite3'
            }

    def __unicode__(self):
        return repr(self)

    def data(self):
        raise Exception("No data method for KeyValue type data.")

    def storage_class_alias(self, file_ext):
        if file_ext == '.sqlite3':
            return 'sqlite3'
        elif file_ext == '.json':
            return 'json'
        else:
            return self.setting('storage-type')

    def value(self, key, throwException=True):
        try:
            return self.storage[key]
        except Exception:
            if throwException:
                raise

    def like(self, key):
        try:
            return self.storage.like(key)
        except AttributeError:
            msg = "The `like()` method is not implemented for storage type '%s'"
            msgargs = self.storage.alias
            raise dexy.exceptions.UserFeedback(msg % msgargs)

    def __getitem__(self, key):
        return self.value(key)

    def append(self, key, value):
        print("appending key %s" % key)
        self.storage.append(key, value)

    def query(self, query):
        return self.storage.query(query)

    def keys(self):
        return self.storage.keys()

    def items(self):
        """
        List of available keys.
        """
        return self.storage.items()

    def save(self):
        try:
            self.storage.persist()
        except Exception as e:
            raise
            msg = "Problem saving '%s': %s" % (self.key, str(e))
            raise dexy.exceptions.InternalDexyProblem(msg)



from dexy.exceptions import UserFeedback
from dexy.exceptions import InternalDexyProblem
from dexy.utils import file_exists
import dexy.exceptions
import dexy.plugin
import os
import shutil
import sqlite3

class Storage(dexy.plugin.Plugin, metaclass=dexy.plugin.PluginMeta):
    """
    Base class for types of Storage.
    """
    _settings = {}

    def assert_location_is_in_project_dir(self, filepath):
        if not self.wrapper.is_location_in_project_dir(filepath):
            msg = "trying to write '%s' outside of '%s'"
            msgargs = (filepath, self.wrapper.project_root,)
            raise dexy.exceptions.UserFeedback(msg % msgargs)

    def __init__(self, storage_key, ext, wrapper):
        self.storage_key = storage_key
        self.ext = ext
        self.wrapper = wrapper
        self._size = None

    def setup(self):
        pass

    def connect(self):
        pass

class GenericStorage(Storage):
    """
    Default type of storage where content is stored in files.
    """
    aliases = ['generic']

    def data_file(self, read=True):
        """
        Location of data file.
        """
        if read:
            if os.path.exists(self.this_data_file()):
                return self.this_data_file()
            elif os.path.exists(self.last_data_file()):
                return self.last_data_file()
            else:
                return self.this_data_file()
        else:
            return self.this_data_file()

    def last_data_file(self):
        """
        Location of data file in last/ cache dir.
        """
        return os.path.join(self.storage_dir(False), "%s%s" % (self.storage_key, self.ext))

    def this_data_file(self):
        """
        Location of data file in this/ cache dir.
        """
        return os.path.join(self.storage_dir(True), "%s%s" % (self.storage_key, self.ext))

    def data_file_exists(self, this):
        if this:
            return os.path.exists(self.this_data_file())
        else:
            return os.path.exists(self.last_data_file())

    def data_file_size(self, this):
        if this:
            return os.path.getsize(self.this_data_file())
        else:
            return os.path.getsize(self.last_data_file())

    def storage_dir(self, this=None):
        if this is None:
            this = (self.wrapper.state in ('walked', 'running'))

        if this:
            cache_dir = self.wrapper.this_cache_dir()
        else:
            cache_dir = self.wrapper.last_cache_dir()
        
        return os.path.join(cache_dir, self.storage_key[0:2])

    def write_data(self, data, filepath=None):
        if not filepath:
            filepath = self.data_file(read=False)

        self.assert_location_is_in_project_dir(filepath)

        if os.path.exists(self.this_data_file()) and not filepath == self.this_data_file():
            shutil.copyfile(self.this_data_file(), filepath)
        else:
             with open(filepath, "wb") as f:
                 if not isinstance(data, str):
                     f.write(data)
                 else:
                     f.write(data.decode("utf-8"))

    def read_data(self):
        with open(self.data_file(read=True), "r") as f:
            return f.read()

    def copy_file(self, filepath):
        """
        If data file exists, copy file and return true. Otherwise return false.
        """
        try:
            self.assert_location_is_in_project_dir(filepath)
            this = (self.wrapper.state in ('walked', 'running', 'ran',))
            shutil.copyfile(self.data_file(this), filepath)
            return True
        except:
            return False

# Sectioned Data
import json
class JsonSectionedStorage(GenericStorage):
    """
    Storage for sectional data using JSON.
    """
    aliases = ['jsonsectioned']

    def read_data(self, this=True):
        with open(self.data_file(this), "r") as f:
            data = json.load(f)
            if hasattr(data, 'keys'):
                msg = "Data storage format has changed. Please clear your dexy cache by running dexy with '-r' option."
                raise UserFeedback(msg)
            return data

    def write_data(self, data, filepath=None):
        if not filepath:
            filepath = self.data_file()

        self.assert_location_is_in_project_dir(filepath)

        with open(filepath, "w") as f:
            json.dump(data, f)

# Key Value Data
class JsonKeyValueStorage(GenericStorage):
    """
    Storage for key value data using JSON.
    """
    aliases = ['json']

    def setup(self):
        self._data = {}

    def append(self, key, value):
        self._data[key] = value

    def keys(self):
        return list(self.data().keys())

    def value(self, key):
        return self.data()[key]

    def __getitem__(self, key):
        return self.value(key)

    def items(self):
        return list(self.data().items())

    def iteritems(self):
        return iter(self.data().items())

    def read_data(self, this=True):
        with open(self.data_file(this), "r") as f:
            return json.load(f)

    def data(self):
        if len(self._data) == 0:
            self._data = self.read_data()
        return self._data

    def persist(self):
        self.write_data(self._data)

    def write_data(self, data, filepath=None):
        if not filepath:
            filepath = self.data_file()

        self.assert_location_is_in_project_dir(filepath)

        with open(filepath, "w") as f:
            json.dump(data, f)

class Sqlite3KeyValueStorage(GenericStorage):
    """
    Storage of key value storage in sqlite3 database files.
    """
    aliases = ['sqlite3']

    def working_file(self):
        sk = self.storage_key[0:2]
        pathargs = (
                self.wrapper.work_cache_dir(),
                sk,
                "%s.sqlite3" % self.storage_key,
                )
        return os.path.join(*pathargs)

    def connect(self):
        self._append_counter = 0
        if self.wrapper.state in ('walked', 'checked', 'running'):
            if file_exists(self.this_data_file()):
                self.connected_to = 'existing'
                self._storage = sqlite3.connect(self.this_data_file())
                self._cursor = self._storage.cursor()
            elif file_exists(self.last_data_file()):
                msg ="Should not only have last data file %s"
                msgargs=(self.last_data_file())
                raise dexy.exceptions.InternalDexyProblem(msg % msgargs)
            else:
                assert not os.path.exists(self.working_file())
                assert os.path.exists(os.path.dirname(self.working_file()))
                self.connected_to = 'working'
                self._storage = sqlite3.connect(self.working_file())
                self._cursor = self._storage.cursor()
                self._cursor.execute("CREATE TABLE kvstore (key TEXT, value TEXT)")
        elif self.wrapper.state == 'walked':
            raise dexy.exceptions.InternalDexyProblem("connect should not be called in 'walked' state")
        else:
            if file_exists(self.last_data_file()):
                self._storage = sqlite3.connect(self.last_data_file())
                self._cursor = self._storage.cursor()
            elif file_exists(self.this_data_file()):
                self._storage = sqlite3.connect(self.this_data_file())
                self._cursor = self._storage.cursor()
            else:
                raise dexy.exceptions.InternalDexyProblem("no data for %s" % self.storage_key)

    def append(self, key, value):
        self._cursor.execute("INSERT INTO kvstore VALUES (?, ?)", (key, value))
        self._append_counter += 1
        if self._append_counter > 1000:
            self.wrapper.log.debug("intermediate commit to sqlite db, resetting append counter to 0")
            self._storage.commit()
            self._append_counter = 0

    def keys(self):
        self._cursor.execute("SELECT key from kvstore")
        return [str(k[0]) for k in self._cursor.fetchall()]

    def items(self):
        self._cursor.execute("SELECT key, value from kvstore")
        for k in self._cursor.fetchall():
            yield (str(k[0]), k[1])

    def value(self, key):
        self._cursor.execute("SELECT value from kvstore where key = ?", (key,))
        row = self._cursor.fetchone()
        if not row:
            raise Exception("No value found for key '%s'" % key)
        else:
            return row[0]

    def like(self, key):
        self._cursor.execute("SELECT value from kvstore where key LIKE ?", (key,))
        row = self._cursor.fetchone()
        if not row:
            raise Exception("No value found for key '%s'" % key)
        else:
            return row[0]

    def query(self, query):
        if not '%' in query:
            query = "%%%s%%" % query
        self._cursor.execute("SELECT * from kvstore where key like ?", (query,))
        return self._cursor.fetchall()

    def __getitem__(self, key):
        return self.value(key)

    def persist(self):
        if self.connected_to == 'existing':
            assert os.path.exists(self.data_file(read=False))
        elif self.connected_to == 'working':
            self.assert_location_is_in_project_dir(self.data_file(read=False))
            self._storage.commit()
            shutil.copyfile(self.working_file(), self.data_file(read=False))
        else:
            msg = "Unexpected 'connected_to' value %s"
            msgargs = self.connected_to
            raise InternalDexyProblem(msg % msgargs)



from dexy.doc import Doc
from tests.utils import assert_in_output
from tests.utils import assert_output
from tests.utils import assert_output_cached
from tests.utils import wrap

def test_pyg4rst():
    o = {}
    o['1'] = ".. code:: python\n\n  print 'hello'"
    assert_output("pyg4rst", "print 'hello'", o, ext=".py")

def test_html():
    assert_in_output("pyg|h", "print 'hello'", """<div class="highlight">""")

def test_png():
    assert_output_cached("pyg|pn", "print 'hello'")

def test_jpg():
    assert_output_cached("pyg|jn", "print 'hello'")

def test_gif():
    assert_output_cached("pyg|gn", "print 'hello'")

def test_pyg4rst_bad_file_extension():
    with wrap() as wrapper:
        wrapper.debug = False
        doc = Doc(
                "hello.xyz|pyg4rst",
                wrapper,
                [],
                contents=" ",
                pyg4rst = { 'allow_unknown_ext' : False }
                )
        wrapper.run_docs(doc)
        assert wrapper.state == 'error'

def test_pygments_bad_file_extension():
    with wrap() as wrapper:
        wrapper.debug = False
        doc = Doc(
                "hello.xyz|pyg",
                wrapper,
                [],
                contents=" ",
                pyg = { 'allow_unknown_ext' : False }
                )
        wrapper.run_docs(doc)
        assert wrapper.state == 'error'

def test_pygments_line_numbering():
    with wrap() as wrapper:
        doc = Doc(
                "hello.py|pyg",
                wrapper,
                [],
                contents=" ",
                pyg = { 'linenos' : True }
                )
        wrapper.run_docs(doc)
        assert "<pre>1</pre>" in str(doc.output_data())

def test_pygments_line_numbering_latex():
    with wrap() as wrapper:
        doc = Doc(
                "hello.py|pyg|l",
                wrapper,
                [],
                contents=" ",
                pyg = { 'linenos' : True }
                )
        wrapper.run_docs(doc)
        assert "firstnumber=1" in str(doc.output_data())

def test_pygments_line_numbering_latex_alt():
    with wrap() as wrapper:
        doc = Doc(
                "hello.py|pyg|l",
                wrapper,
                [],
                contents=" ",
                pyg = { 'line-numbers' : True }
                )
        wrapper.run_docs(doc)
        assert "firstnumber=1" in str(doc.output_data())



from dexy.commands.utils import template_text
from pygments import highlight
from pygments.lexers import PythonLexer
import dexy.filter
import inspect
import pygments.formatters

extra_nodoc_aliases = ('-',)

def filters_command(
        alias="", # Print docs for this filter.
        example=False, # Whether to run included examples (slower).
        nocolor=False, # Skip syntax highlighting if showing source code.
        source=False, # Print source code of filter.
        versions=False # Print the installed version of external software (slower).
        ):
    """
    Prints list of available filters or docs for a particular filter.
    """
    if alias:
        help_for_filter(alias, example, source, nocolor)
    else:
        list_filters(versions)

def help_for_filter(alias, run_example, show_source, nocolor):
    instance = dexy.filter.Filter.create_instance(alias)

    print('')
    print(instance.setting('help'))

    print('')
    print("aliases: %s" % ", ".join(instance.setting('aliases')))
    print("tags: %s" % ", ".join(instance.setting('tags')))
    print('')

    print("Converts from file formats:")
    for ext in instance.setting('input-extensions'):
        print("   %s" % ext)
    print('')

    print("Converts to file formats:")
    for ext in instance.setting('output-extensions'):
        print("   %s" % ext)
    print('')

    print('Settings:')
    for k in sorted(instance._instance_settings):
        if k in dexy.filter.Filter.nodoc_settings:
            continue
        if k in ('aliases', 'tags'):
            continue

        tup = instance._instance_settings[k]
        print("    %s" % k)

        for line in inspect.cleandoc(tup[0]).splitlines():
            print("        %s" % line)

        print("        default value: %s" % tup[1])
        print('')

    examples = instance.setting('examples')
    example_templates = {}
    for alias in examples:
        try:
            template_instance = dexy.template.Template.create_instance(alias)
            example_templates[alias] = template_instance
        except dexy.exceptions.InactivePlugin:
            pass

    if examples:
        print('')
        print("Examples for this filter:")
        for alias, template in example_templates.items():
            print('')
            print("  %s" % alias)
            print("            %s" % inspect.getdoc(template.__class__))

        if run_example:
            for alias, template in example_templates.items():
                print('')
                print('')
                print("Running example: %s" % template.setting('help'))
                print('')
                print('')
                print(template_text(template))

    print('')
    print("For online docs see http://dexy.it/filters/%s" % alias)
    print('')
    print("If you have suggestions or feedback about this filter,")
    print("please contact info@dexy.it")
    print('')

    if show_source:
        print('')
        source_code = inspect.getsource(instance.__class__)
        if nocolor:
            print(source_code)
        else:
            formatter = pygments.formatters.TerminalFormatter()
            lexer = PythonLexer()
            print(highlight(source_code, lexer, formatter))

def list_filters(versions):
        print("Installed filters:")
        for filter_instance in dexy.filter.Filter:
            # Should we show this filter?
            no_aliases = not filter_instance.setting('aliases')
            no_doc = filter_instance.setting('nodoc')
            not_dexy = not filter_instance.__class__.__module__.startswith("dexy.")
            exclude = filter_instance.alias in extra_nodoc_aliases

            if no_aliases or no_doc or not_dexy or exclude:
                continue

            # generate version message
            if versions:
                if hasattr(filter_instance, 'version'):
                    version = filter_instance.version()
                    if version:
                        version_message = "Installed version: %s" % version
                    else:
                        msg = "'%s' failed, filter may not be available."
                        msgargs = filter_instance.version_command()
                        version_message = msg % msgargs
                else:
                    version_message = ""


            filter_help = "  " + filter_instance.alias + \
                    " : " + filter_instance.setting('help').splitlines()[0]

            if versions and version_message:
                filter_help += " %s" % version_message

            print(filter_help)

        print('')
        print("For more information about a particular filter,")
        print("use the -alias flag and specify the filter alias.")
        print('')




from tests.utils import assert_in_output
from tests.utils import assert_output
from tests.utils import assert_output_matches
import inspect
import os

def test_node():
    assert_output("nodejs", "console.log('hello');", "hello\n")

def test_rd():
    rd = """
     \\name{load}
     \\alias{load}
     \\title{Reload Saved Datasets}
     \description{
       Reload the datasets written to a file with the function
       \code{save}.
     }
    """
    expected = "Reload the datasets written to a file with the function \u2018save\u2019."
    assert_in_output('rdconv', rd, expected, ext=".Rd")

def test_redcloth():
    expected = "<p>hello <strong>bold</strong></p>" + os.linesep
    assert_output("redcloth", "hello *bold*", expected)

def test_redclothl():
    expected = "hello \\textbf{bold}" + os.linesep + os.linesep
    assert_output("redclothl", "hello *bold*", expected)

def test_lynxdump():
    assert_output_matches('lynxdump', "<p>hello</p>", "\s*hello\s*", ext=".html")

def test_strings():
    assert_output('strings', "hello\bmore", "hello\nmore\n")

def test_php():
    php = inspect.cleandoc("""<?php
    echo(1+1);
    ?>""")
    assert_output('php', php, "2")

def test_ragel_ruby_dot():
    ragel = inspect.cleandoc("""
        %%{
          machine hello_and_welcome;
          main := ( 'h' @ { puts "hello world!" }
                  | 'w' @ { puts "welcome" }
                  )*;
        }%%
          data = 'whwwwwhw'
          %% write data;
          %% write init;
          %% write exec;
        """)
    assert_in_output('ragelrubydot', ragel, "digraph hello_and_welcome", ext=".rl")

def test_python():
    assert_output('py', 'print(1+1)', "2" + os.linesep)

def test_bash():
    assert_output('bash', 'echo "hello"', "hello\n")

def test_rhino():
    assert_output('rhino', "print(6*7)", "42\n")

def test_cowsay():
    assert_in_output('cowsay', 'hello', 'hello')

def test_cowthink():
    assert_in_output('cowthink', 'hello', 'hello')

def test_figlet():
    assert_in_output('figlet', 'hello', "| |__   ___| | | ___  ")

def test_man_page():
    assert_in_output('man', 'ls', 'list directory contents')

def test_ruby():
    assert_output('rb', 'puts "hello"', "hello\n")

def test_sloccount():
    assert_in_output('sloccount', 'puts "hello"', "ruby=1", ext=".rb")

def test_irb_subprocess_stdout_filter():
    assert_in_output('irbout', 'puts "hello"', '> puts "hello"')

def test_lua():
    assert_output('lua', 'print ("Hello")', "Hello\n")

def test_wiki2beamer():
    wiki = inspect.cleandoc("""==== A simple frame ====
    * with a funky
    * bullet list
    *# and two
    *# numbered sub-items
    """)

    assert_in_output('wiki2beamer', wiki, "\\begin{frame}")
    assert_in_output('wiki2beamer', wiki, "\\begin{enumerate}")
